#!/usr/bin/python

import errno
import os
import sys
import fcntl
import datetime
import urllib2
import base64
import json
import time
import subprocess
import commands
import hashlib
import traceback
import ssl
import signal
from threading import Lock, Thread
from multiprocessing import Process, Pipe, current_process
from StringIO import StringIO

from optparse import OptionParser
from xml.etree import ElementTree
from xml.etree.ElementTree import Element, SubElement, Comment, tostring
import xml.etree.ElementTree as ET
from xml.dom import minidom
from subprocess import Popen, PIPE

SUCCESS = 0
FAILURE = -1
ONE_MB = (1024*1024)

g_info_file_path = "/var/lib/aprotect_teardown_info"
g_info_lock_file_path = "/var/lib/aprotect_teardown_info.lck"
g_lock_f = None
error_msg = None
g_disk_warning_msg = None
g_http_status_code = 0
gServer5xxRetryCount = 5
gTaskWaitTime = 60
gTaskRetryCount = 5
gIscsiCmdRetryCount = 5
teardown_commands = []
gTerminate = 0
gVerbose = 1
gTraceLevel = 1
gMsgLock = None

# Define some global constants that will be used mainly when
# making the REST calls
CONST_HTTPS = "https"
CONST_PORT = "9440"
CONST_API_PATH = "api/nutanix"
CONST_V2 = "v2.0"
CONST_V3 = "v3"

CONST_CLUSTER = "cluster"
CONST_HOSTS = "hosts"
CONST_VMS = "vms"
CONST_GET_VM_DISKS = "include_vm_disk_config=true"
CONST_GET_VM_NICS = "include_vm_nic_config=true"
CONST_VMS_AND_DISKS = CONST_VMS + "/?" + CONST_GET_VM_DISKS
CONST_VMS_DISKS_ATTACH = "disks/attach"
CONST_STORAGE_CONTAINERS = "storage_containers"
CONST_IDEMPOTENCE_ID = "idempotence_identifiers"
CONST_VM_SNAPSHOTS = "vm_snapshots"
CONST_DATA_CHANGED_REGIONS = "data/changed_regions"
CONST_VOLUME_GROUPS = "volume_groups"
CONST_TASKS = "tasks"
CONST_NETWORKS = "networks"
CONST_NICS = "nics"

CONST_APPLICATION_CONSISTENT = 'APPLICATION_CONSISTENT'
CONST_CRASH_CONSISTENT = 'CRASH_CONSISTENT'


# Some HTTP error codes. Can add more as we move forward
HTTP_BAD_REQUEST = 400
HTTP_UNAUTHORIZED = 401
HTTP_FORBIDDEN = 403
HTTP_NOT_FOUND  = 404
HTTP_INTERNAL_SERVER_ERROR = 500

DISK_NAME_PREFIX = "Disk"

# Build all command line options along with their help text
parser = OptionParser()
parser.add_option("-s", "--server", dest="server",
	help="IP Address for AHV Cluster")
parser.add_option("-u", "--username", dest="username",
	help="Username for AHV Cluster")
parser.add_option("-p", "--password", dest="password",
	help="Password for AHV Cluster")
parser.add_option("-b", "--bpcmd", dest="bpcmd",
	help="Get information for - Cluster, VMs or Storage Containers")
parser.add_option("-U", "--uuid", dest="uuid",
	help="UUID of the VM for which the operation needs to be performed")
parser.add_option("-d", "--disklist", dest="disklist", action="store_true",
	help="Disk listing for a single VM")
parser.add_option("-B", "--backup", action="store_true",
	help="Backup VM")
parser.add_option("-P", "--postbackup", action="store_true",
	help="PostBackup VM")
parser.add_option("-S", "--starttime", dest="start_time",
	help="Backup start_time")
parser.add_option("-E", "--exclude", dest="exclude",
	help="Exclude disks from backup")
parser.add_option("-q", "--quiesce", dest="quiesce",
	help="Quiesce")

parser.add_option("-n", "--snapshot", dest="snapshot",
	help="Snapshot uuid from prior previous backup")
parser.add_option("-X", "--delsnapshot", dest="delsnapshot",
	help="Snapshot uuid to delete")
parser.add_option("-r", "--keepsnapshot", dest="keepsnapshot",
	help="Snapshot uuid to keep")

parser.add_option("-H", "--hardwarehash", dest="hardwarehash",
	help="Hardware hash from prior previous backup")

parser.add_option("-R", "--restore", action="store_true",
	help="Restore VM")

parser.add_option("-O", "--postrestore", action="store_true",
	help="PostRestore VM")

parser.add_option("-N", "--newvmname", dest="newvmname",
	help="new name for VM")

parser.add_option("-C", "--storagecontainer", dest="storagecontainer",
	help="Storage container")

parser.add_option("-m", "--metadata", dest="metadata",
	help="restore metadata of excluded disks")

parser.add_option("-x", "--xml_filename", dest="xml_filename",
	help="metadata xml filename")

parser.add_option("-g", "--volumegroup", dest="volumegroup",
	help="volume group uuid")

parser.add_option("-j", "--jobid", dest="jobid",
	help="job id")

parser.add_option("-J", "--ppid", dest="ppid",
	help="parent pid")

parser.add_option("-L", "--lstart", dest="lstart",
	help="parent pid lstart")

parser.add_option("-I", "--cleanupidle", action="store_true",
	help="cleanup idle info files")

parser.add_option("-c", "--cli", action="store_true",
	help="command line interface")

parser.add_option("-v", "--verbose", action="store_true",
	help="verbose")

parser.add_option("-t", "--prmtest", dest="prmtest",
	help="prmtest")

(options, args) = parser.parse_args()

def sig_handler(signum, frame):
	global gTerminate
	gTerminate += 1
	dbg_msg( 'Signal handler called with signal %s' % signum)
	return

# Helper function to print all log messages to STDERR
def dbg_msg(msg):
	if gVerbose == 0:
		return

	if gMsgLock:
		gMsgLock.acquire()

	try:
		curr_time = datetime.datetime.now()
		# TODO : The in-built way of using 'tzinfo()' doesn't print anything.
		# Need to see if there's anything else that can be done to print the timezone.
		# For now, keeping it as "+0000".
		formatted_time = curr_time.strftime("%Y-%m-%d %H:%M:%S.000 +0000")
		name = current_process().name
		if name == "MainProcess":
			print >> sys.stderr, formatted_time,': D',msg
		else:
			print >> sys.stderr, formatted_time,name,': D',msg

	except Exception as e:
		print >> sys.stderr, ': E %s' % (e)

	if gMsgLock:
		gMsgLock.release()

# Return a pretty-printed XML string for the Element.
# TODO : Remove this if not needed
def prettify(elem):

	rough_string = ElementTree.tostring(elem, 'utf-8')
	reparsed = minidom.parseString(rough_string)
	return reparsed.toprettyxml(indent=" ")

# Helper function to create an appropriately indented XML
def indent(elem, level = 0):

	i = "\n" + level*"  "
	if len(elem):
		if not elem.text or not elem.text.strip():
			elem.text = i + "  "
		if not elem.tail or not elem.tail.strip():
			elem.tail = i
		for elem in elem:
			indent(elem, level + 1)
		if not elem.tail or not elem.tail.strip():
			elem.tail = i
	else:
		if level and (not elem.tail or not elem.tail.strip()):
			elem.tail = i

def SubElementWithText(parent, tag, text):
	child = SubElement(parent, tag)
	child.text = text
	return child

def SubElementWithTextFromNamedItem(parent, key, dictionary):
	tag = key
	data = dictionary[key];
	if isinstance(data, basestring):
		text = data
	else:
		text = str(data)
	child = SubElement(parent, tag)
	child.text = text
	return child

def SubElementFromData(parent, tag, data):
	s = SubElement(parent, tag)
	buildxmlFromData(s, data)
	return s

def SubElementFromList(parent, tag, data, childtag):
	listhead = SubElement(parent, tag)
	if isinstance(data, tuple) or isinstance(data, list):
		for v in data:
			s = SubElement(listhead, childtag)
			buildxmlFromData(s, v)
	return listhead

def buildxmlFromData(parent, data):
	if isinstance(data, dict):
		for k, v in data.iteritems():
			s = SubElement(parent, k)
			buildxmlFromData(s, v)
	elif isinstance(data, tuple) or isinstance(data, list):
		for v in data:
			s = SubElement(parent, 'i')
			buildxmlFromData(s, v)
	elif isinstance(data, basestring):
		parent.text = data
	else:
		parent.text = str(data)
	return parent


def data2xml(d, name='data'):
    r = et.Element(name)
    return et.tostring(buildxml(r, d))

def buildxml(r, d):
    if isinstance(d, dict):
        for k, v in d.iteritems():
            s = et.SubElement(r, k)
            buildxml(s, v)
    elif isinstance(d, tuple) or isinstance(d, list):
        for v in d:
            s = et.SubElement(r, 'i')
            buildxml(s, v)
    elif isinstance(d, basestring):
        r.text = d
    else:
        r.text = str(d)
    return r

def mkDataFromElementTree( tree ):
	print "%s : %s : %s\n" % (tree.tag, tree.attrib, tree.text)
	for child in tree:  
		dbg_msg("%s : %s\n" % (child.tag, child.text) )
		print "%s : %s" % (child.tag, child.text)


def get_et_boolean(child):
	if child.text in [ "1", "True" ]:
		return True
	else:
		return False

def add_xml_promote_next_with_reason(parent, reason):
	child = SubElementWithText(parent, 'promote_next_backup', '1')
	child = SubElementWithText(parent, 'promote_next_backup_reason', reason)

# Puts the 'error_msg' as a part of the XML being returned.
# Prints the XML on STDOUT as well.
def output_xml_for_error(error_msg):

	root_aprotect = Element('aprotect')

	child_aprotect = SubElement(root_aprotect, 'aprotect_error')

	child_aprotect_error_msg = SubElement(child_aprotect, 'error_message')
	child_aprotect_error_msg.text = error_msg

	indent(root_aprotect)
	ElementTree.dump(root_aprotect)

def output_xml_for_error_with_promote_next(error_msg, reason):

	root_aprotect = Element('aprotect')

	child_aprotect = SubElement(root_aprotect, 'aprotect_error')

	child_aprotect_error_msg = SubElement(child_aprotect, 'error_message')
	child_aprotect_error_msg.text = error_msg

	add_xml_promote_next_with_reason(root_aprotect, reason)

	indent(root_aprotect)
	ElementTree.dump(root_aprotect)

def get_special_error_msg(e):

	error_msg = "Internal server error occurred"
	try:
		failure_response = e.read()
		#dbg_msg("ERROR response:\n" + failure_response)
		failure_result = json.loads(failure_response)
		#if 'detailed_message' in failure_result:
		#	detailed_message = failure_result['detailed_message']
		#	dbg_msg( detailed_message )
		json_print = json.dumps(failure_result, indent = 4, separators = (',', ': '))
		dbg_msg("########### JSON output for failure scenario ###########")
		dbg_msg("\n" + json_print)
		dbg_msg("Message from REST API: %s" % failure_result['message'])
		error_msg = failure_result['message']
	except:
		# Generic
		dbg_msg("Unable to find details about error.")

	return error_msg

def append_error_msg(error_msg, text):
	if error_msg == None:
		error_msg = text
	else:
		if (len(error_msg) > 0 and error_msg[-1] != '\n'):
			error_msg += '\n'
		error_msg += text
	return error_msg

# Helper function that converts a given HTTP error code to an appropriate
# error message.
# For now, the common error codes are handled. For other error codes,
# a generic error is returned. However, before returning the generic error,
# the actual error message is logged for debugging purposes.
# TODO: Add other HTTP error codes as needed
def get_text_for_http_error(e):

	dbg_msg("http error: %s" % (e))

	if (e.code == HTTP_BAD_REQUEST):
		error_msg = "Incorrect format of the request to get information from the Nutanix Cluster."
	elif (e.code == HTTP_UNAUTHORIZED):
		error_msg = "Invalid set of credentials entered for the Nutanix Cluster. Please verify your credentials have administrative privileges."
	elif (e.code == HTTP_FORBIDDEN):
		error_msg = "Request cannot be handled."
	elif (e.code == HTTP_NOT_FOUND):
		#error_msg = "Request for this operation cannot be fulfilled by the Nutanix Cluster."
		error_msg = "Not found. Request for this operation cannot be fulfilled by the Nutanix Cluster."
	elif (e.code == HTTP_INTERNAL_SERVER_ERROR):
		# Can occur when a request is sent to the Cluster that it can't understand
		# TODO: Improve the below snippet
		#if (options.disklist == 1):
			error_msg = get_special_error_msg(e)
		#else:
		#	error_msg = "Internal server error occurred"
	else:
		reason = ""
		if hasattr(e,'reason'):
			reason = e.reason
		dbg_msg("Unknown error %s occurred: %s" % (e.code, reason))
		error_msg = "Unknown error occurred."

	return error_msg

# Helper function that takes the input as an URL, runs it and returns the REST response from the same.
def get_rest_api_response(url,body=None,method=None):

	#dbg_msg("Entering get_rest_api_response for URL: %s" % url)
	dbg_msg("Entering get_rest_api_response")

	result = None
	error_msg = None

	password_decoded = base64.b64decode(options.password);
	# TODO : Remove below log message
	# dbg_msg("Decoded password is %s" % password_decoded);

	# When using Basic Authorization for the REST request, the 'base64' form of the pair -
	#	username:password
	# is required to be put in the header
	username_password = options.username + ":" + password_decoded

	authKey = base64.b64encode(username_password)
	headers = {"Content-Type":"application/json",
			"Authorization":"Basic " +
			authKey}

	if body:	# and self.content_type == "application/json":
		#print "Data: %s" % body
		#body = json.dumps(body)
		body = json.dumps(body, indent = 4, separators = (',', ': '))
		#dbg_msg("POST Data: %s" % body)
		request = urllib2.Request(url, body)
		request.get_method = lambda: "POST"
	else:
		request = urllib2.Request(url)

	if method:
		request.get_method = lambda: method

	dbg_msg("%s %s" % (request.get_method(), url) )
	if body:
		dbg_msg("POST Data: %s" % body)

	# Can the below fail?
	#request = urllib2.Request(url)

	for key,value in headers.items():
		request.add_header(key,value)

	retry_count = gServer5xxRetryCount
	if (retry_count < 0):
		retry_count = 0
	#dbg_msg("Server5xxRetryCount: %d" % retry_count)

	max_tries = 1 + retry_count
	for count in range(0, max_tries):
		if (count > 0):
			delaysec = 1 + count*3
			dbg_msg("Delay %d sec, then retry request" % delaysec)
			time.sleep( delaysec )  # seconds
		g_http_status_code = 0
		error_msg = None
		try:
			response = urllib2.urlopen(request)
			g_http_status_code = response.getcode() #response.status
		except urllib2.HTTPError, e:
			g_http_status_code = e.code
			error_msg = get_text_for_http_error(e)
			# check for errors we recognize
			ix = error_msg.find('Unknown vgs:')
			if ix > -1:
				# Couldn't find the volume group. no point in retrying.
				return (result, error_msg)
	 	except Exception as e:
            		dbg_msg("Error: %s" % e)
			if 'Connection timed out' in str(e):
				error_msg = "Connection timed out when trying to get information from Nutanix.";
			else:
				error_msg = "Unknown error occurred when trying to get information from Nutanix.";
			dbg_msg(error_msg)
			return (result, error_msg)

		if g_http_status_code >= 500 and g_http_status_code < 600:
			# response code 5xx is server error.
			# let's delay then retry.
			dbg_msg("Server error")
		else:
			# no retry, so break out of loop
			break

	if error_msg:
		return (result, error_msg)
			

	#print "########### Headers are: ###########\n%s" % response.info().headers
	#print "########### Read response: ###########\n%s" % response.read()

	data = response.read()
	#print data

	# 'data' is in a JSON format
	# Print the complete JSON response in the logs for reference purpose.
	# It will be a huge chunk of data. Can revisit this later (to think of
	# any other idea), if needed.
	# Check for data from transforming. DELETE will not have response data.
	if data != None and len(data) > 0:
		result = json.loads(data)
		if not 'changed_regions' in url or gTraceLevel >= 3:
			json_print = json.dumps(result, indent = 4, separators = (',', ': '))
			dbg_msg("########### JSON output from REST API ###########")
			dbg_msg("\n" + json_print)

	#dbg_msg("Leaving get_rest_api_response for URL: %s" % url)
	dbg_msg("Leaving get_rest_api_response")

	return (result, error_msg)

def gen_url_ntnx_api_v2():
	url = CONST_HTTPS + "://" + options.server + ":" + CONST_PORT + "/" + \
		CONST_API_PATH + "/" + CONST_V2
	return url

def gen_url_ntnx_api_v3():
	url = CONST_HTTPS + "://" + options.server + ":" + CONST_PORT + "/" + \
		CONST_API_PATH + "/" + CONST_V3
	return url


def get_server_credentials(server):
	# queries:
	query = "select username,password,domain from bp.client_credentials where credential_id=" + \
		"(select credential_id from bp.nodes where node_no=" + \
			"(select node_no from bp.ahv_clusters where address='%s'))" % server
	cmd = 'psql -At -Upostgres bpdb -c "%s"' % query
	lines, error_msg = do_cmd_expect_output(cmd, hideStdout=True)
	if error_msg != None:
		if len(error_msg) == 0:
			error_msg = "Server credentials not found."
		return None, error_msg
	if lines == None or len(lines) == 0:
		error_msg = "Server credentials not found."
		return None, error_msg
	fields = lines[0].split('|')
	username = fields[0]
	password = fields[1]
	domain = fields[2]
	dbg_msg("username: %s" % username )
	# TODO: remove later
	##dbg_msg("password: %s" % password )
	dbg_msg("password: %s" % "*" )
	dbg_msg("domain  : %s" % domain )
	return [username, password, domain], None

def get_lock():
	global g_lock_f
	try:
		if g_lock_f == None:
			g_lock_f = os.open(g_info_lock_file_path, os.O_WRONLY | os.O_CREAT )
			#g_lock_f = os.open(g_lock_file_path, "rw")
			#g_lock_f = open(g_lock_file_path, "rw")
			fcntl.lockf(g_lock_f, fcntl.LOCK_EX)
	except Exception as e:
		dbg_msg( "get_lock: %s" % e)
		return -1
	return 0

def release_lock():
	global g_lock_f
	if g_lock_f:
		os.close( g_lock_f )  #g_lock_f.close()
		g_lock_f = None
	return 0

def write_info_file():
	global teardown_commands
	dbg_msg("write_info_file: %s" % (teardown_commands))
	filename = g_info_file_path ###+ "." + options.ppid
	try:
		f = open(filename, "w")
		#fcntl.lockf(f, fcntl.LOCK_EX)
	except Exception as e:
		dbg_msg(e)
	if f:
		for line in teardown_commands:
			f.write( line )
			if line[-1] != '\n':
				f.write('\n')
		f.close()

def read_info_file():
	global teardown_commands
	filename = g_info_file_path
	teardown_commands = []
	f = None
	try:
		f = open(filename, "r")
		#fcntl.lockf(f, fcntl.LOCK_EX)
	except Exception as e:
		dbg_msg(e)
	if f:
		teardown_commands = [ line.rstrip('\n') for line in f.readlines() ]
		f.close()
	dbg_msg("read_info_file: %s" % (teardown_commands))


def get_teardown_id():
	ppid = os.getppid()
	lstart = ""

	if options.ppid != None:
		ppid = int( options.ppid )
	if options.lstart != None:
		lstart = options.lstart
	if len(lstart) == 0 and ppid > 0:
		lines, error_msg = do_cmd_expect_output( "ps -p %d -o lstart=" % ppid )
		if lines and len(lines) > 0 and error_msg == None:
			lstart = lines[0].rstrip('\n')
	return "%s|%s" % (str(ppid), lstart)
	
	
def add_teardown_entry(line):
	global teardown_commands
	oldmask = os.umask(077)
	get_lock()
	#time.sleep( 10 )  # seconds
	read_info_file()
	# inseart at head so that teardown is done in reverse order
	teardown_commands.insert(0, line)
	write_info_file()
	release_lock()

def remove_teardown_entry( entry ):
	global teardown_commands
	oldmask = os.umask(077)
	get_lock()
	read_info_file()
	if entry in teardown_commands:
		teardown_commands.remove(entry)
	if 0:
		found = -1
		for ix in range(0,len(teardown_commands)):
			line = teardown_commands[ix]
			dbg_msg("remove_teardown_entry: line %s  entry %s" % (line,entry))
			if line == entry:
				dbg_msg("remove_teardown_entry: matched %d" % ix)
				found = ix
				break
		dbg_msg("remove_teardown_entry: found %d" % found)
		if found != -1:
			teardown_commands.remove(ix)

	write_info_file()
	release_lock()


def add_teardown_entry_ss_uuid( ss_uuid ):
	# columns: ver|ppid|lstart|tries|ss_remove|server|username|password|ss_uuid
	add_teardown_entry( "0|" + get_teardown_id() + "|0|ss_remove|%s|%s|%s|%s" % 
				(options.server, "u", "p", ss_uuid) )

def remove_teardown_entry_ss_uuid( ss_uuid ):
	# columns: ver|ppid|lstart|tries|ss_remove|server|username|password|ss_uuid
	remove_teardown_entry( "0|" + get_teardown_id() + "|0|ss_remove|%s|%s|%s|%s" % 
				(options.server, "u", "p", ss_uuid) )


def add_teardown_entry_vg_uuid( vg_uuid ):
	# columns: ver|ppid|lstart|tries|vg_remove|server|username|password|vg_uuid
	add_teardown_entry( "0|" + get_teardown_id() + "|0|vg_remove|%s|%s|%s|%s" % 
				(options.server, "u", "p", vg_uuid) )
				#(options.server, options.username, options.password, vg_uuid) )

def remove_teardown_entry_vg_uuid( vg_uuid ):
	# columns: ver|ppid|lstart|tries|vg_remove|server|username|password|vg_uuid
	remove_teardown_entry( "0|" + get_teardown_id() + "|0|vg_remove|%s|%s|%s|%s" % 
				(options.server, "u", "p", vg_uuid) )


def add_teardown_entry_iscsi( target_ip_addr, vm_uuid ):
	# columns: ver|ppid|lstart|tries|isci_detach|target_ip_addr|vm_uuid
	add_teardown_entry( "0|" + get_teardown_id() + "|0|iscsi_detach|%s|%s" % (target_ip_addr, vm_uuid))

def remove_teardown_entry_iscsi( target_ip_addr, vm_uuid ):
	# columns: ver|ppid|lstart|tries|isci_detach|target_ip_addr|vm_uuid
	remove_teardown_entry( "0|" + get_teardown_id() + "|0|iscsi_detach|%s|%s" % (target_ip_addr, vm_uuid))


def add_teardown_entry_vm_restoring( vm_uuid ):
	# columns: ver|ppid|lstart|tries|restoring_vm|vm_uuid
	add_teardown_entry( "0|" + get_teardown_id() + "|0|vm_restoring|%s" % (vm_uuid))

def remove_teardown_entry_vm_restoring( vm_uuid ):
	# columns: ver|ppid|lstart|tries|restoring_vm|vm_uuid
	remove_teardown_entry( "0|" + get_teardown_id() + "|0|vm_restoring|%s" % (vm_uuid))


def process_idle_info_files():
	global teardown_commands
	oldmask = os.umask(077)
	get_lock()
	read_info_file()
	release_lock()

	max_num_tries = 336  # 336 == 24*14 == hourly for 14 days

	# get dpu ip addresses
	(client_ip_addr_list, error_msg) = get_dpu_ip_addresses()
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't get appliance IP address.")
		output_xml_for_error(error_msg)
		sys.exit(2)
	dbg_msg("dpu ip_addrs: %s" % (client_ip_addr_list,) )

	##ppid = options.ppid
	##lstart = options.lstart
	to_remove = []
	pidlstart_running_map = {}

	# since we can modify the teardown_commands list while scanning it, make a tmp copy 
	tmp_teardown_commands = list(teardown_commands)
	for ix in range(0,len(tmp_teardown_commands)):

		line = tmp_teardown_commands[ix]
		dbg_msg("process_info_file: line: %s" % line )

		is_successful_action = 0

		fields = line.split('|')

		# columns
		# ver|ppid|lstart|op|...
		version = fields[0]
		pid = fields[1]
		lstart = fields[2]

		# is process running
		process_lookup_str = pid + "|" + lstart
		if process_lookup_str in pidlstart_running_map:
			is_running = pidlstart_running_map[ process_lookup_str ]
		else:
			is_running = is_process_running(pid, lstart)
			pidlstart_running_map[ process_lookup_str ] = is_running

		# skip if the pid is still running
		if is_running:
			continue

		# the pid is no longer running
		# 
		##if ppid != fields[1] or lstart != fields[2]:
		##	continue 
		num_tries = int( fields[3] )
		op = fields[4]

		if op == 'iscsi_detach':
			# columns
			# ...|isci_detach|target_ip_addr|vm_uuid
			dbg_msg("op: iscsi_detach")
			target_ip_addr = fields[5]
			vm_uuid = fields[6]
			(result, error_msg) = iscsi_detach(target_ip_addr, vm_uuid)
			if (error_msg != None):
				dbg_msg(error_msg)
			else:
				is_successful_action = 1

		elif op == 'vg_remove':
			# columns
			# ...|vg_remove|server|username|password|vg_uuid
			dbg_msg("op: vg_remove")
			err = 0
			options.server = fields[5]
			username = fields[6]
			password = fields[7]
			domain = None
			if username == 'u' or password == 'p':
				# lookup credentials from database
				result, error_msg = get_server_credentials(options.server)
				if (error_msg != None):
					dbg_msg("Server credentials not found.")
					err = 1
				else:
					username, password, domain = result
			if err == 0:
				if domain != None and len(domain) > 0:
					options.username = username + '@' + domain
				else:
					options.username = username
				options.password = password
				vg_uuid = fields[8]
				dbg_msg("vg_uuid: " + vg_uuid)
				dbg_msg("Deleting volume group vg_uuid=%s" % (vg_uuid))
				(result, error_msg) = cleanup_vg_by_vg_uuid( vg_uuid )
				###TEST
				###error_msg = "ERROR TEST"
				if (error_msg != None):
					dbg_msg(error_msg)
				else:
					is_successful_action = 1
		elif op == 'ss_remove':
			# columns
			# ...|ss_remove|server|username|password|ss_uuid
			dbg_msg("op: ss_remove")
			err = 0
			options.server = fields[5]
			username = fields[6]
			password = fields[7]
			domain = None
			if username == 'u' or password == 'p':
				# lookup credentials from database
				result, error_msg = get_server_credentials(options.server)
				if (error_msg != None):
					dbg_msg("Server credentials not found.")
					err = 1
				else:
					username, password, domain = result
			if err == 0:
				if domain != None and len(domain) > 0:
					options.username = username + '@' + domain
				else:
					options.username = username
				options.password = password
				ss_uuid = fields[8]
				dbg_msg("ss_uuid: " + ss_uuid)
				dbg_msg("Deleting snapshot ss_uuid=%s" % (ss_uuid))
				(result, error_msg) = cleanup_snapshot_by_ss_uuid(ss_uuid)
				if (error_msg != None):
					dbg_msg(error_msg)
				else:
					is_successful_action = 1

		elif op == 'vm_restoring':
			dbg_msg("op: vm_restoring")
			vm_uuid = fields[5]
			# no action to take; just remove entry, which indicated restore in progress
			is_successful_action = 1

		else:
			dbg_msg("unknown teardown op: " + op)

		# if action was successful, then remove entry
		# else update try count in the entry
		if is_successful_action:
			dbg_msg("teardown action successful: %s" % line)
			remove_teardown_entry( line )
		else:
			dbg_msg("teardown action failed: %s" % line)
			# if too many tries, then remove entry
			# else update try count
			remove_teardown_entry( line )
			num_tries += 1
			if (num_tries < max_num_tries):
				fields[3] = str(num_tries)
				newentry = "|".join(fields)
				add_teardown_entry( newentry )

def get_running_uni_process_restoring_vm_list():
	oldmask = os.umask(077)
	get_lock()
	read_info_file()
	release_lock()

	pidlstart_running_map = {}
	vm_uuid_list = []

	for ix in range(0,len(teardown_commands)):
		line = teardown_commands[ix]
		#dbg_msg("is_vm_in_active_uni_process: line: %s" % line )

		fields = line.split('|')

		# columns: ver|ppid|lstart|op|...
		version = fields[0]
		pid = fields[1]
		lstart = fields[2]
		num_tries = int( fields[3] )
		op = fields[4]

		# is process running
		is_running = False
		process_lookup_str = pid + "|" + lstart
		if process_lookup_str in pidlstart_running_map:
			is_running = pidlstart_running_map[ process_lookup_str ]
		else:
			is_running = is_process_running(pid, lstart)
			pidlstart_running_map[ process_lookup_str ] = is_running

		if op == 'vm_restoring':
			#dbg_msg("op: vm_restoring")
			vm_uuid = fields[5]
			vm_uuid_list.append( vm_uuid )

	return vm_uuid_list

def is_running_uni_process_restoring_vm(search_uuid):
	vm_uuid_list = get_running_uni_process_restoring_vm_list()
	if vm_uuid_list and (search_uuid in vm_uuid_list):
		return True
	return False

def is_process_running(pid, lstart):
	is_running = 1
	if os.path.exists("/proc/%s" % pid):
		dbg_msg("process is still running: %s" % pid)
		# is it the same one
		lines, error_msg = do_cmd_expect_output( "ps -p %s -o lstart=" % pid )
		if lines and len(lines) > 0 and error_msg == None:
			curlstart = lines[0]
			if curlstart != lstart:
				# it's a different process
				dbg_msg("it's a different process")
				is_running = 0
			
	else:
		dbg_msg("process is not running: %s" % pid)
		is_running = 0
	return is_running


def OLD_process_idle_info_files():
	global teardown_commands
	filename = g_info_file_path
	f = None
	try:
		f = open(filename, "r")
	except Exception as e:
		dbg_msg(e)
	if f:
		lines = [ line.rstrip('\n') for line in f.readlines() ]
		f.close()
	dbg_msg("process_idle_info_files: lines: %s" % (lines))
	
	# check pid and time to see if it is running
	for line in lines:

		fields = line.split('|')
		version = fields[0]
		pid = fields[1]
		lstart = fields[2]

		# is process running?
		is_running = 1
		if os.path.exists("/proc/%s" % pid):
			dbg_msg("process is still running: %s" % pid)
			# is it the same one
			lines, error_msg = do_cmd_expect_output( "ps -p %s -o lstart=" % pid )
			if lines and len(lines) > 0 and error_msg == None:
				curlstart = lines[0]
				if curlstart != lstart:
					# it's a different process
					is_running = 0
				
		else:
			dbg_msg("process is not running: %s" % pid)
			is_running = 0

		# cleanup?
		if not is_running:
			process_info_file( pid, lstart)

def get_proc_meminfo_committed_free():
	commit_limit = 0
	committed_AS = 0
	committed_free = 0
	f = None

	try:
		f = open('/proc/meminfo')
		for line in f.readlines():
			line = line.strip()
			parts = line.split()
			#print "parts: %s" % (parts)
			if parts[0] == 'CommitLimit:':
				commit_limit = int( parts[1] )
			elif parts[0] == 'Committed_AS:':
				committed_AS = int( parts[1] )

		f.close()
		f = None

		if commit_limit > 0 and committed_AS > 0:
			committed_free = commit_limit - committed_AS

		dbg_msg( "CommitLimit   : %12d" % (commit_limit) )
		dbg_msg( "Committed_AS  : %12d" % (committed_AS) )
		dbg_msg( "committed free: %12d" % (committed_free) )

		# committed_free is in KB so adjust it
		if committed_free > 0:
			committed_free *= 1024

	except OSError as e:
		dbg_msg("get_proc_meminfo_committed_free: OS error: %s" % (e))
	except Exception as e:
		dbg_msg("get_proc_meminfo_committed_free: Unexpected error: %s" % (e))
	finally:
		if f:
			f.close()

	return committed_free

def get_multiprocess_max_procs():
	one_gb = 1024*1024*1024
	max_procs = 1
	max_workers = ini_get_int("aprotect", "MaxWorkersPerBackup", -1)
	dbg_msg("max_workers: %d" % max_workers)
	if max_workers >= 0:
		max_procs = max_workers
	else:
		# auto config max procs
		committed_free = get_proc_meminfo_committed_free()
		if (committed_free > 0):
			#if committed_free > (4 * one_gb):
			#	max_procs = 8
			#elif committed_free > (2 * one_gb):
			#	max_procs = 4
			#elif committed_free > one_gb:
			#	max_procs = 2
			x = float(committed_free) / float(one_gb)
			if x > 1.0:
				max_procs = int( x * 2.0 )
				if max_procs > 16:
					max_procs = 16
	dbg_msg("get_multiprocess_max_procs: return %d" % max_procs)
	return max_procs

# Helper function to get the socket count of the COMPLETE cluster i.e.,
# sum of the socket counts of the individual nodes.
# Note that this does an additional REST API call to retrieve that information.
def get_cluster_socket_count():

	dbg_msg("Entering get_cluster_socket_count")

	cluster_socket_count = 0
	error_msg = None

	# Build the URL to get the appropriate information
	get_hosts_url = gen_url_ntnx_api_v2() + "/" + CONST_HOSTS

	(result, error_msg) = get_rest_api_response(get_hosts_url)
	if (error_msg != None):
		return (cluster_socket_count, error_msg)
	else:
		dbg_msg("Successfully retrieved REST API response for URL: %s" % get_hosts_url)

	metadata = result.get('metadata')
	if metadata == None:
		return (cluster_socket_count, "Failed to get node metadata from the Nutanix cluster.")

	number_of_nodes = metadata.get('total_entities')
	if number_of_nodes == None:
		return (cluster_socket_count, "Failed to get the node count from the Nutanix cluster.")

	dbg_msg("Total number of nodes: %s" % number_of_nodes)

	for i in range(0,number_of_nodes):
		try:
			current_node_socket_count = result['entities'][i]['num_cpu_sockets']
		except:
			dbg_msg("Can't retrieve socket count for node %s" % i)
			continue

		# check for non null value for this field
		if current_node_socket_count != None:
			cluster_socket_count = cluster_socket_count + current_node_socket_count

	dbg_msg("Cluster total socket count: %s" % cluster_socket_count)

	dbg_msg("Leaving get_cluster_socket_count")

	return (cluster_socket_count, error_msg)

# Main function to traverse over the response received from the REST API that
# was called to get all the Nutanix AHV cluster information.
# The needed information will be put into the 'child_aprotect' XML and returned.
def get_cluster_information(child_aprotect, result):

	dbg_msg("Entering get_cluster_information")

	if ( (result['name'] == None) or
		(result['uuid'] == None) or
		(result['version'] == None) ):
		error_msg = "One or more of the required fields for Nutanix Acropolis " \
				"hypervisor registration is not available."
		return error_msg

	(cluster_socket_count, error_msg) = get_cluster_socket_count()
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't get Nutanix cluster information.")
		return error_msg
	else:
		dbg_msg("Successfully retrieved socket count for the cluster")

	child_aprotect_name = SubElement(child_aprotect, 'name')
	child_aprotect_name.text = result['name']

	child_aprotect_cluster_uuid = SubElement(child_aprotect, 'ahv_cluster_uuid')
	child_aprotect_cluster_uuid.text = result['uuid']

	child_aprotect_version = SubElement(child_aprotect, 'version')
	child_aprotect_version.text = result['version']

	child_aprotect_socket = SubElement(child_aprotect, 'num_cpu_sockets')
	child_aprotect_socket.text = str(cluster_socket_count)

	dbg_msg("Leaving get_cluster_information")

def ntnx_get_cluster_info():

	dbg_msg("Entering ntnx_get_cluster_info")

	error_msg = None

	# Build the URL to get the appropriate information
	get_cluster_url = gen_url_ntnx_api_v2() + "/" + CONST_CLUSTER

	(result, error_msg) = get_rest_api_response(get_cluster_url)
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't get Nutanix cluster information.")
		return (None, error_msg)

	dbg_msg("Leaving ntnx_get_cluster_info")

	return (result, error_msg)

def ntnx_get_cluster_data_ip_addr():

	dbg_msg("Entering ntnx_get_cluster_data_ip_addr")

	no_data_ip_error_msg = "Couldn't get cluster iSCSI data services IP address.  Check that the iSCSI data services IP address is configured for the Nutanix cluster."
	error_msg = None

	(cluster_info, error_msg) = ntnx_get_cluster_info()
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't get Nutanix cluster information.")
		return (None, error_msg)

	if 'cluster_external_data_services_ipaddress' not in cluster_info:
		error_msg = no_data_ip_error_msg
		return (None, error_msg)

	data_ip_addr = cluster_info['cluster_external_data_services_ipaddress']
	if data_ip_addr == None or (len(data_ip_addr) == 0):
		error_msg = no_data_ip_error_msg
		return (None, error_msg)

	dbg_msg("cluster data ip addr: %s" % data_ip_addr)

	dbg_msg("Leaving ntnx_get_cluster_data_ip_addr")
	return (data_ip_addr, error_msg)

def ntnx_get_network_info(network_uuid):
	error_msg = None
	url = gen_url_ntnx_api_v2() + "/" + CONST_NETWORKS + "/" + network_uuid
	(result, error_msg) = get_rest_api_response(url)
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't get Nutanix cluster network information.")
		return (None, error_msg)
	return (result, error_msg)

def gen_disk_name(index):
	# Since Nutanix doesn't give a name to VMs' disks, we create
	# a generic name of the form "Disk 0", "Disk 1", etc.
	disk_name = DISK_NAME_PREFIX + "-" + str(index)
	return disk_name

def get_vm_disk_cdrom_uuid_list(vm_disk_info):

	list = []

	number_of_disks = len(vm_disk_info)
	for j in range(0,number_of_disks):
		try:
			# a cdrom without a disk will not have a vmdisk_uuid
			vmdisk_uuid = vm_disk_info[j]['disk_address']['vmdisk_uuid']
			#disk_size = vm_disk_info[j]['size']
			#disk_device_index = vm_disk_info[j]['disk_address']['device_index']
			#disk_label = vm_disk_info[j]['disk_address']['disk_label']
			#is_empty = vm_disk_info[j]['is_empty']

			is_cdrom = vm_disk_info[j]['is_cdrom']
			if (is_cdrom):
				list.append( vmdisk_uuid )
		except:
			dbg_msg("One or more parameters unavailable for disk %s" % j)
			continue

	return list

def get_vm_disk_information(vm_disk_info, vm_xml, useCdrom=False):
	global g_disk_warning_msg

	number_of_disks = len(vm_disk_info)
	dbg_msg("This VM has %s disks" % number_of_disks)

	# prepare exclude list
	exclude_list = []
	if options.exclude != None and len(options.exclude) > 0 and options.exclude != 'empty':
		exclude_list = options.exclude.split(':')

	disks_xml = SubElement(vm_xml, 'disks')

	naming_index = 0

	for j in range(0,number_of_disks):
		try:
			# cdrom should not be in disk list, but needs to be in vm.
			is_cdrom = vm_disk_info[j]['is_cdrom']
			is_empty = vm_disk_info[j]['is_empty']
			
			disk_address = vm_disk_info[j]['disk_address']
			disk_label = disk_address['disk_label']

			vmdisk_uuid = ""
			if 'vmdisk_uuid' in disk_address:
				vmdisk_uuid = disk_address['vmdisk_uuid']

			disk_size = 0
			if 'size' in vm_disk_info[j]:
				disk_size = vm_disk_info[j]['size']

			if is_cdrom and not useCdrom:
				naming_index += 1
				continue

			# skip disks which are attached volume groups
			if 'volume_group_uuid' in disk_address:
				naming_index += 1
				#volume_group_uuid vm_disk_info[j].find['volume_group_uuid']
				msg = "WARNING: Skipping disk '%s' attached to volume group." % (disk_label)
				g_disk_warning_msg = append_error_msg(g_disk_warning_msg, msg)
				continue

	 	except Exception as e:
			dbg_msg("%s" % (e))
			dbg_msg("One or more parameters unavailable for disk %s" % j)
			continue

		# Since Nutanix doesn't give a name to VMs' disks, we create
		# a generic name of the form "Disk 0", "Disk 1", etc.

		# use disk label for disk name
		disk_name = disk_label

		tagname = 'disk'
		# give cdrom a tag name of 'cdrom' instead of 'disk'
		if is_cdrom:
			tagname = 'cdrom'

		disk = SubElement(disks_xml, tagname)
		SubElementWithText(disk, 'key', str(naming_index))
		SubElementWithText(disk, 'name', disk_name)
		SubElementWithText(disk, 'label', disk_label)
		dbg_msg("\tSuccessfully added disk name %s" % disk_name)
		if not is_empty:
			disk_uuid = SubElement(disk, 'uuid')
			disk_uuid.text = vmdisk_uuid
			dbg_msg("\tSuccessfully added disk UUID %s" % vmdisk_uuid)
			disk_capacity = SubElement(disk, 'capacity')
			disk_capacity.text = str(disk_size)
			dbg_msg("\tSuccessfully added disk size %d" % disk_size)

		####SubElementFromData(disk, 'boot', vm_info['boot'] )
		buildxmlFromData(disk, vm_disk_info[j] )

		if vmdisk_uuid in exclude_list:
			dbg_msg("is_excluded disk: %s" % vmdisk_uuid)
			SubElementWithText(disk, 'is_excluded', '1')

		naming_index += 1


# Main function to traverse over the response received from the REST API that
# was called to get all the VM information.
# The needed information will be put into the 'child_aprotect' XML and returned.
# Depending on the 'get_info' parameter value, the disk information for each VM
# may or may not be returned.
def get_vm_information(get_info, child_aprotect, result):

	dbg_msg("Entering get_vm_information")

	try:
		number_of_vms = result['metadata']['count']
	except:
		dbg_msg("Unable to get list of VMs")
		error_msg = "Failed to get list of virtual machines from Nutanix cluster"
		return error_msg

	dbg_msg("There are %s virtual machines in the inventory" % number_of_vms)

	# get list of VMs with a restore in progress
	vm_uuid_list = get_running_uni_process_restoring_vm_list()

	child_aprotect_vms = SubElement(child_aprotect, 'virtual_machines')

	for i in range(0,number_of_vms):

		try:
			vm_name = result['entities'][i]['name']
			vm_uuid = result['entities'][i]['uuid']
			entity = result['entities'][i]
		except:
			dbg_msg("Name and/or UUID is not defined for VM %s. Skipping it." % i)
			continue			

		# don't return VMs with a restore in progress
		if vm_uuid_list and (vm_uuid in vm_uuid_list):
			dbg_msg("Restore is in progress for VM %s. Skipping it." % vm_uuid)
			continue

		child_aprotect_vm = SubElement(child_aprotect_vms, 'virtual_machine')
		child_aprotect_vm_name = SubElementWithText(child_aprotect_vm, 'name', vm_name)
		child_aprotect_vm_uuid = SubElementWithText(child_aprotect_vm, 'uuid', vm_uuid)

		if (get_info == CONST_VMS_AND_DISKS):

			try:
				vm_disk_info = result['entities'][i]['vm_disk_info']
			except:
				dbg_msg("Unable to retrieved disk information for VM %s" \
						% result['entities'][i]['name'])
				continue

			get_vm_disk_information(vm_disk_info, child_aprotect_vm)

	dbg_msg("Leaving get_vm_information")
	return None

def get_storage_container_information(get_info, child_aprotect, result):

	dbg_msg("Entering get_storage_container_information")

	number_of_containers = 0
	count = 0
	error_msg = None

	storage_containers_xml = SubElement(child_aprotect, 'storage_containers')

	number_of_containers = result['metadata']['total_entities']
	dbg_msg("Total number of containers: %s" % number_of_containers)

	for i in range(0,number_of_containers):
		try:
			entity = result['entities'][i]
			storage_container_uuid = entity['storage_container_uuid']
			name = entity['name']
			max_capacity = entity['max_capacity']

			# these field values match the values shown in the nutanix ui for the containers
			# ui             ==> usage_stats field
			# free (logical) ==> "storage.user_free_bytes"
			# max_capactiy   ==> "storage.user_capacity_bytes"

			usage_stats = entity['usage_stats']
			user_capacity_bytes = usage_stats['storage.user_capacity_bytes']
			user_free_bytes = usage_stats['storage.user_free_bytes']

			container_xml = SubElement(storage_containers_xml, 'storage_container')
			SubElementWithText( container_xml, 'storage_container_uuid', storage_container_uuid )
			SubElementWithText( container_xml, 'name', name )
			SubElementWithText( container_xml, 'user_capacity_bytes', str(user_capacity_bytes) )
			SubElementWithText( container_xml, 'user_free_bytes', str(user_free_bytes) )

		except:
			dbg_msg("Can't retrieve container fields for index %d" % i)
			error_msg = "Can't retrieve container fields for index %d\n" % i
			continue

	dbg_msg("Leaving get_storage_container_information")

	return error_msg

def get_xml_for_bpcmd():

	dbg_msg("Entering get_xml_for_bpcmd")

	if (options.bpcmd == "get_cluster_info"):
		get_info = CONST_CLUSTER
	elif (options.bpcmd == "get_vm_info"):
		get_info = CONST_VMS
	elif (options.bpcmd == "get_vm_and_disk_info"):
		get_info = CONST_VMS_AND_DISKS
	elif (options.bpcmd == "get_storage_containers_info"):
		get_info = CONST_STORAGE_CONTAINERS
	else:
		error_msg = "Invalid value for bpcmd."
		return error_msg

	# Build the URL to get the appropriate information
	get_info_url = gen_url_ntnx_api_v2() + "/" + get_info

	(result, error_msg) = get_rest_api_response(get_info_url)
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't get Nutanix cluster information.")
		return error_msg
	else:
		dbg_msg("Successfully retrieved REST API response for URL: %s" % get_info_url)

	root_aprotect = Element('aprotect')
	child_aprotect = SubElement(root_aprotect, 'aprotect_info')

	if (get_info == CONST_CLUSTER):
		error_msg = get_cluster_information(child_aprotect, result)
	elif ( (get_info == CONST_VMS) or (get_info == CONST_VMS_AND_DISKS) ):
		error_msg = get_vm_information(get_info, child_aprotect, result)
	elif (get_info == CONST_STORAGE_CONTAINERS):
		error_msg = get_storage_container_information(get_info, child_aprotect, result)

	if (error_msg != None):
		return error_msg
	else:
		dbg_msg("Successfully retrieved %s information" % get_info)

	# Prints all the XML on one line
	#print tostring(root_aprotect)

	# Prints XML in a little weird format
	#print prettify(root_aprotect)
	#root = ElementTree.parse(root_aprotect).getroot()

	# The below 2 statements will print the XML in more readable form
	indent(root_aprotect)
	ElementTree.dump(root_aprotect)

	dbg_msg("Leaving get_xml_for_bpcmd")


def get_disk_info_for_vm():

	dbg_msg("Entering get_disk_info_for_vm")

	# Build the URL to get the appropriate information
	get_vm_info_url = gen_url_ntnx_api_v2() + "/" + CONST_VMS + "/" + \
			options.uuid + "/?" + CONST_GET_VM_DISKS

	(result, error_msg) = get_rest_api_response(get_vm_info_url)
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't get VM disk information.")
		return error_msg
	else:
		dbg_msg("Successfully retrieved REST API response for URL: %s" % get_vm_info_url)

	root_aprotect = Element('aprotect')
	child_aprotect = SubElement(root_aprotect, 'aprotect_info')

	virtual_machine_xml = SubElement(child_aprotect, 'virtual_machine')

	try:
		vm_name = result['name']
		vm_uuid = result['uuid']
	except:
		dbg_msg("Name and/or UUID is not defined for VM %s." % options.uuid)
		error_msg = "Virtual machine's information needed to get disk information is unavailable"
		return error_msg

	vm_name_xml = SubElement(virtual_machine_xml, 'name')
	vm_name_xml.text = vm_name
	vm_uuid_xml = SubElement(virtual_machine_xml, 'uuid')
	vm_uuid_xml.text = vm_uuid

	if 'vm_disk_info' in result:
		vm_disk_info = result['vm_disk_info']
		get_vm_disk_information(vm_disk_info, virtual_machine_xml)
	else:
		dbg_msg("Unable to retrieve disk information for VM %s" % result['name'])

	indent(root_aprotect)
	ElementTree.dump(root_aprotect)

	dbg_msg("Leaving get_disk_info_for_vm")

def get_xml_for_vm_info_for_metadata():

	dbg_msg("Entering get_vm_info_for_metadata")

	# Build the URL to get the appropriate information
	get_vm_info_url = gen_url_ntnx_api_v2() + "/" + CONST_VMS + "/" + \
		options.uuid + "/?" + CONST_GET_VM_DISKS + "&" + CONST_GET_VM_NICS

	(vm_info, error_msg) = get_rest_api_response(get_vm_info_url)
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't get VM information.")
		return None, None, error_msg
	else:
		dbg_msg("Successfully retrieved REST API response for URL: %s" % get_vm_info_url)

	virtual_machine_xml = Element('virtual_machine')

	try:
		vm_name = vm_info['name']
		vm_uuid = vm_info['uuid']
	except:
		dbg_msg("Name and/or UUID is not defined for VM %s." % options.uuid)
		error_msg = "Virtual machine's information needed to get disk information is unavailable"
		return None, None, error_msg

	vm_name_xml = SubElementWithText(virtual_machine_xml, 'name', vm_name)
	vm_uuid_xml = SubElementWithText(virtual_machine_xml, 'uuid', vm_uuid)

	for key in ['num_cores_per_vcpu', 'num_vcpus', 'gpus_assigned', 'memory_mb', 'power_state', 'ha_priority',
			'allow_live_migrate', 'timezone', 'vm_logical_timestamp' ]:
			##'host_uuid' ]:
		SubElementWithTextFromNamedItem(virtual_machine_xml, key, vm_info)

	if 'boot' in vm_info:
		SubElementFromData(virtual_machine_xml, 'boot', vm_info['boot'] )

	if 'vm_disk_info' in vm_info:
		vm_disk_info = vm_info['vm_disk_info']
		get_vm_disk_information(vm_disk_info, virtual_machine_xml, True)
	else:
		dbg_msg("Unable to retrieve disk information for VM %s" % vm_info['name'])

	if 'vm_nics' in vm_info:
		# get network info to get vlan id.  we'll use vlan_id during restore.
		for vm_nic in vm_info['vm_nics']:
			network_uuid = vm_nic['network_uuid']
			result, error_msg = ntnx_get_network_info(network_uuid)
			if error_msg != None:
				dbg_msg("Couldn't get network info for network_uuid: %s" % network_uuid)
				# 
			else:
				vlan_id = result['vlan_id']
				# augment vm_nic info with the vlan_id
				vm_nic['vlan_id'] = vlan_id

		SubElementFromList(virtual_machine_xml, 'vm_nics', vm_info['vm_nics'], 'vm_nic' )
			

	#indent(root_aprotect)
	#ElementTree.dump(root_aprotect)

	dbg_msg("Leaving get_disk_info_for_vm")
	return vm_info, virtual_machine_xml, None

def gen_hardware_hash_string(vm_info):
	hw_hash = None
	error_msg = None
	number_of_disks = 0

	if not 'vm_disk_info' in vm_info:
		dbg_msg("This VM has %d disks" % number_of_disks)
		# use the uuid as hw string
		hw_string = options.uuid;
		hw_hash_string = hashlib.md5(hw_string).hexdigest()
		dbg_msg("gen_hardware_hash_string:")
		dbg_msg("    hw_string      : %s" % hw_string)
		dbg_msg("    hw_hash_string : %s" % hw_hash_string)
		return hw_hash_string, error_msg


	vm_disk_info = vm_info['vm_disk_info']
	number_of_disks = len(vm_disk_info)
	dbg_msg("This VM has %d disks" % number_of_disks)

	disk_label_to_disk_info = {}

	for j in range(0,number_of_disks):
		try:
			is_cdrom = vm_disk_info[j]['is_cdrom']
			if is_cdrom:
				continue
			vmdisk_uuid = vm_disk_info[j]['disk_address']['vmdisk_uuid']
			disk_size = vm_disk_info[j]['size']
			disk_label = vm_disk_info[j]['disk_address']['disk_label']
			disk_device_index = vm_disk_info[j]['disk_address']['device_index']

			disk_label_to_disk_info[ disk_label ] = {
				'disk_label': disk_label,
				'vmdisk_uuid': vmdisk_uuid,
				'size': disk_size
				}
		except:
			dbg_msg("One or more parameters unavailable for disk %s" % j)
			continue

	hw_string = ""

	# accumulate disks to string, ordered by disk label.
	#  hw_string += disk_label:disk_uuid ...
	for key in sorted(disk_label_to_disk_info.iterkeys()):
		di = disk_label_to_disk_info[key]
		if len(hw_string) > 0:
			hw_string += ","
		hw_string += di['disk_label'] + ':' + di['vmdisk_uuid'] + ':' + str(di['size'])

	hw_hash_string = hashlib.md5(hw_string).hexdigest()
	dbg_msg("gen_hardware_hash_string:")
	dbg_msg("    hw_string      : %s" % hw_string)
	dbg_msg("    hw_hash_string : %s" % hw_hash_string)

	return hw_hash_string, error_msg

def get_total_change_size(crt_info, skip_zeros):
	sum = 0
	try:
		file_size = crt_info['file_size']
		if 'regions' in crt_info: 
			regions = crt_info['regions']		# api v3 seen in aos 5.1.0.1
		else:
			regions = crt_info['region_list'] 	# api v3 spec seen in aos 5.1.4.0
		regions_len = len(regions)
		for j in range(0,regions_len):
			region = regions[j]
			offset = region['offset']
			length = region['length']
			type = region['type']
			if skip_zeros and type[0] == 'Z':
				continue
			sum += length
	except:
		dbg_msg("Could not get crt info")
		return None, "Could not get crt info"

	return sum, None

def get_xml_for_crt_info(crt_info, skip_zeros, crt_info_xml):

	try:
		file_size = crt_info['file_size']
		if 'regions' in crt_info:
			regions = crt_info['regions']		# api v3 seen in in aos 5.1.0.1
		else:
			regions = crt_info['region_list']	# api v3 spec seen in aos 5.1.4.0
	except:
		dbg_msg("Could not get crt info")
		return None, "Could not get crt info"

	if crt_info_xml == None:
		crt_info_xml = Element('crt_info')
		file_size_xml = SubElement(crt_info_xml, 'file_size')
		file_size_xml.text = str(file_size)

	regions_len = len(regions)

	regions_xml = crt_info_xml.find('regions')
	if regions_xml == None:
		regions_xml = SubElement(crt_info_xml, 'regions')

	skip_count = 0

	for j in range(0,regions_len):
		try:
			region = regions[j]
			offset = region['offset']
			length = region['length']
			type = region['type']
		except:
			dbg_msg("One or parameters unavailable for region %d" % j)
			continue

		if skip_zeros and type[0] == 'Z':
			skip_count += 1
			continue

		region = SubElement(regions_xml, 'region')

		region_offset = SubElement(region, 'offset')
		region_offset.text = str(offset)

		region_type = SubElement(region, 'type')
		region_type.text = type

		region_length = SubElement(region, 'length')
		region_length.text = str(length)

	prev_regions_len = 0
	num_regions_xml = crt_info_xml.find('num_regions')
	if num_regions_xml == None:
		num_regions_xml = SubElement(crt_info_xml, 'num_regions')
	else:
		prev_regions_len = int(num_regions_xml.text)
	num_regions_xml.text = str(prev_regions_len + regions_len - skip_count)

	# The below 2 statements will print the XML in more readable form
	#indent(root_xml)
	#ElementTree.dump(root_xml)

	return crt_info_xml, None

def get_xml_for_aligned_result( result, is_skip_zeros ):

	file_size = result['file_size']
	count = result['count']
	offsets = result['offsets']
	lengths = result['lengths']
	types = result['types']


	crt_info_xml = Element('crt_info')
	file_size_xml = SubElementWithText(crt_info_xml, 'file_size', str(file_size) )
	regions_xml = SubElement(crt_info_xml, 'regions')

	skip_count = 0
	regions_len = count

	dbg_msg( "=== aligned and consolidated changed regions" )
	for j in range(0,regions_len):

		offset = offsets[j]
		length = lengths[j]
		type_str = types[j]
		dbg_msg( "=== offset: 0x%08x, length: 0x%08x, type: %s" % (offset, length, type_str))

		if is_skip_zeros and type_str[0] == 'Z':
			skip_count += 1
			continue

		region = SubElement(regions_xml, 'region')
		region_offset = SubElementWithText(region, 'offset', str(offset) )
		region_type = SubElementWithText(region, 'type', str(type_str) )
		region_length = SubElementWithText(region, 'length', str(length) )

	num_regions_xml = SubElement(crt_info_xml, 'num_regions')
	num_regions_xml.text = str(regions_len - skip_count)

	# The below 2 statements will print the XML in more readable form
	#print "===TESTING=========================================="
	#indent(crt_info_xml)
	#ElementTree.dump(crt_info_xml)
	#print "===================================================="

	return crt_info_xml, None

def align_changed_regions(crt_info_list, is_skip_zeros):
	align_size = 1024*1024
	offsets = []
	lengths = []
	types = []
	curr_index = -1

	file_size = 0

	list_len = len(crt_info_list)
	for crt_info in crt_info_list:

		file_size = crt_info['file_size']
		if 'regions' in crt_info:
			regions = crt_info['regions']		# api v3 seen in aos 5.1.0.1
		else:
			regions = crt_info['region_list']	# api v3 spec seen in aos 5.1.4.0
		regions_len = len(regions)
		for j in range(0, regions_len):
			region = regions[j]
			offset = region['offset']
			length = region['length']
			type_str = region['type']

			###for testing
			###dbg_msg("CRT offset: %d, length: %d, type: %s" % (offset, length, type_str) )

			if is_skip_zeros and type_str[0] == 'Z':
				continue

			lower = (offset / align_size) * align_size
			end = offset + length
			upper = ((end + (align_size - 1)) / align_size) * align_size

			if upper > file_size:
				upper = file_size

			if curr_index == -1:
				# first block
				curr_index = 0
				offsets.append( lower )
				lengths.append( upper - lower )
				types.append( type_str )
			else:
				curr_upper = offsets[curr_index] + lengths[curr_index]
				if (lower > (curr_upper + align_size)):
					# new block
					curr_index += 1
					offsets.append( lower )
					lengths.append( upper - lower )
					types.append( type_str )
				else:
					# if (lower >= offsets[curr_index] and (lower <= (offsets[curr_index] + lengths[curr_index])):
					# contiguous with current block
					if upper > curr_upper:
						# expand block
						curr_upper = upper
						lengths[curr_index] = upper - offsets[curr_index]
						# update type_str if needed
						if types[curr_index][0] == 'Z' and type_str[0] != 'Z':
							types[curr_index] = type_str

	result = {
		'file_size': file_size,
		'count': curr_index + 1,
		'offsets': offsets,
		'lengths': lengths,
		'types': types }
	return result

def ntnx_get_idempotence_id(client_identifier):

	dbg_msg("Entering ntnx_get_idempotence_id")

	# Build the URL to get the appropriate information
	url = gen_url_ntnx_api_v3() + "/" + CONST_IDEMPOTENCE_ID + "/" + client_identifier

	(result, error_msg) = get_rest_api_response(url)
	if (error_msg != None):
		return None, error_msg
	else:
		dbg_msg("Successfully retrieved response for URL: %s" % url)

	try:
		client_identifier = result['client_identifier']
		# count
		uuid = result['uuid_list'][0]
	except:
		error_msg = "Could not parse idempotence id response."
		return (0, error_msg)

	dbg_msg("Leaving ntnx_get_idempotence_id")
	return (uuid, error_msg)

def ntnx_create_idempotence_id(client_identifier):

	dbg_msg("Entering ntnx_create_idempotence_id")

	# Build the URL to get the appropriate information
	url = gen_url_ntnx_api_v3() + "/" + CONST_IDEMPOTENCE_ID

	body = {
		'client_identifier': client_identifier,
		'count': 1
	}

	(result, error_msg) = get_rest_api_response(url, body)
	if (error_msg != None):
		return (None, error_msg)
	else:
		dbg_msg("Successfully retrieved response for URL: %s" % url)

	try:
		client_identifier = result['client_identifier']
		uuid = result['uuid_list'][0]
	except:
		error_msg = "Could not parse idempotence id response."
		return (None, error_msg)

	dbg_msg("Leaving ntnx_create_idempotence_id")
	return (uuid, error_msg)

def ntnx_delete_idempotence_id(client_identifier):
	dbg_msg("Entering ntnx_delete_idempotence_id")

	# Build the URL to get the appropriate information
	url = gen_url_ntnx_api_v3() + "/" + CONST_IDEMPOTENCE_ID + "/" + client_identifier

	(result, error_msg) = get_rest_api_response(url, None, "DELETE")
	if (error_msg != None):
		return 0, error_msg
	else:
		dbg_msg("Successfully retrieved REST API response for URL: %s" % url)

	dbg_msg("Leaving ntnx_delete_idempotence_id")
	return (0, error_msg)

def ntnx_get_task(task_uuid):
	# get task details
	dbg_msg("Get task details, task_uuid %s" % task_uuid)
	url = gen_url_ntnx_api_v2() + "/" + CONST_TASKS + "/" + task_uuid
	(result, error_msg) = get_rest_api_response(url,None,None)
	return result, error_msg

def ntnx_is_result_complete(task_uuid):
	# These tasks can be in one of the following states:
	# 'QUEUED','RUNNING', 'SUCCEEDED', 'ABORTED', 'FAILED'.
	(result, error_msg) = ntnx_get_task(task_uuid)
	#if result and str(status) == "200":
	progress_status = None
	if result and error_msg == None:
		percent_complete = result.get('percentage_complete')
		progress_status = result.get('progress_status')
		if (str(percent_complete) == "100" and progress_status.upper() == "SUCCEEDED"):
			entity = json.dumps(result.get('entity_list')[-1])
			entity_list = json.loads(entity)
			#return entity_list.get('entity_id'), error_msg
			return result, error_msg

		elif progress_status.upper() in ["FAILED", "ABORTED"]:
			meta_response = result.get('meta_response')
			error_detail = meta_response.get('error_detail')
			error_msg = error_detail
			return result, error_msg

	return result, error_msg

def ntnx_track_completion_status(result):
	#max_tries = 30
	wait_time = 3 # seconds
	uuid = None
	error_msg = None
	progress_status = None
	task_uuid = result.get('task_uuid')
	if not task_uuid:
		error_msg = "Couldn't get task_uuid"
		return None, error_msg

	t0 = time.time()
	max_wait_sec = gTaskWaitTime
	if (max_wait_sec < 30):
		max_wait_sec = 30
	dbg_msg("max_wait_sec: %d" % max_wait_sec)

	count = 0
	while not error_msg:
		count += 1
		tdiff = time.time() - t0
		if tdiff >= max_wait_sec:  # if (count > max_tries):
			error_msg = "Timeout waiting for task to complete."
			return None, error_msg

		# sleep the first iteration to give the server a chance to complete task
		dbg_msg("delay %d sec" % wait_time)
		time.sleep(wait_time)

		(result, error_msg) = ntnx_is_result_complete(task_uuid)
		if error_msg:
			dbg_msg("Error getting task result: %s" % error_msg)

		elif result and 'progress_status' in result:
			# task states: 'QUEUED','RUNNING', 'SUCCEEDED', 'ABORTED', 'FAILED'.
			progress_status = result.get('progress_status')
			if progress_status and progress_status.upper() not in ["QUEUED", "RUNNING"]:
				break

	if error_msg:
		dbg_msg( "task status error detail:" + error_msg)
		return None, error_msg

	return result, error_msg

def ntnx_ss_list(vm_uuid):
	if vm_uuid:
		dbg_msg("Entering ntnx_ss_list(vm_uuid=%s)" % vm_uuid)
	else:
		dbg_msg("Entering ntnx_ss_list(vm_uuid=None)")

	url = gen_url_ntnx_api_v3() + "/" + CONST_VM_SNAPSHOTS + "/" + "list"

	if (vm_uuid == None):
		body = {
			'filter': 'entity_uuid==*',
			'kind': 'vm_snapshot'
		}
	else:
		body = {
			'filter': 'entity_uuid==%s' % vm_uuid,
			'kind': 'vm_snapshot'
		}

	(result, error_msg) = get_rest_api_response(url, body)
	if (error_msg != None):
		return (None, error_msg)

	dbg_msg("Leaving ntnx_ss_list")
	return (result, error_msg)

def ntnx_ss_get(ss_uuid):
	dbg_msg("Entering ntnx_ss_get( %s)" % ss_uuid)

	# Build the URL to get the appropriate information
	url = gen_url_ntnx_api_v3() + "/" + CONST_VM_SNAPSHOTS + "/" + ss_uuid

	(result, error_msg) = get_rest_api_response(url)
	if (error_msg != None):
		return (None, error_msg)

	dbg_msg("Leaving ntnx_ss_get")
	return (result, error_msg)

def ntnx_ss_delete(ss_uuid):
	dbg_msg("Entering ntnx_ss_delete(%s)" % ss_uuid)

	# Build the URL to get the appropriate information
	url = gen_url_ntnx_api_v3() + "/" + CONST_VM_SNAPSHOTS + "/" + ss_uuid

	(result, error_msg) = get_rest_api_response(url, None, "DELETE")
	if (error_msg != None):
		return (1, error_msg)
	else:
		dbg_msg("Successfully retrieved REST API response for URL: %s" % url)

	dbg_msg("Leaving ntnx_ss_delete")
	return (0, None)

def get_snapshot_type_message(ss_info):
	warnmsg = ""
	msg = ""
	snapshot_type = ""
	try:
		snapshot_type = ss_info['status']['snapshot_type']
		spec_snapshot_type = ss_info['spec']['snapshot_type']
	except:
		dbg_msg("Couldn't get snapshot_type from ss_info")

	if snapshot_type == CONST_APPLICATION_CONSISTENT:
		msg = "Snapshot for this Backup was created with application-consistent setting."

	elif snapshot_type == CONST_CRASH_CONSISTENT:
		if spec_snapshot_type == CONST_APPLICATION_CONSISTENT:
			warnmsg = "The backup requested an application-consistent snapshot, but a crash-consistent snapshot was created."
			#warnmsg = "This backup was performed with a snapshot without I/O quiescing i.e., " + \
                        #        "the snapshot was performed with crash-consistent setting."
		else:
			msg = "Snapshot for this Backup was created with crash-consistent setting."

	dbg_msg("get_snapshot_type_message: returning msg=%s, warnmsg=%s" % (msg, warnmsg) )

	return msg, warnmsg

def ntnx_ss_create(vm_uuid, ss_name, ss_uuid):
	dbg_msg("Entering ntnx_ss_create(%s)" % ss_uuid)

	url = gen_url_ntnx_api_v3() + "/" + CONST_VM_SNAPSHOTS

	snapshot_type = CONST_APPLICATION_CONSISTENT
	if options.quiesce != None:
		if  options.quiesce.upper() == 'NO':
			snapshot_type = CONST_CRASH_CONSISTENT

	body = {
		"spec": {
			"resources": {
				"entity_uuid": vm_uuid
			},
			"snapshot_type": snapshot_type,
			"name": ss_name,
		},
		"api_version": "3.0",
		"metadata": {
			"kind": "vm_snapshot",
			"uuid": ss_uuid
		}
	}

	(result, error_msg) = get_rest_api_response(url, body)
	if (error_msg != None):
		dbg_msg("ntnx_ss_create: %s" % error_msg)
		return (None, error_msg)

	# v3.0 intent states:
	# An intent can be in any one of the following states:
	# 'PENDING', 'RUNNING', 'COMPLETE',
	# 'ERROR', 'DELETE_PENDING', 'RETRY', 'DELETE_IN_PROGRESS'.

	# wait for completion
	t0 = time.time()
	delay_list = [ 1, 2, 3, 5, 10, 15 ]

	# get snapshot timeout value from master.ini
	max_wait_sec = ini_get_int("aprotect", "SnapshotWaitTime", 600)
	if max_wait_sec < 30:
		max_wait_sec = 30
	dbg_msg("max_wait_sec: %d" % max_wait_sec)

	error_msg = None
	ss_state = "kPending"

	count = 0
	while not error_msg and (ss_state not in ["kComplete", "COMPLETE"] ):
		# check wait time
		tdiff = time.time() - t0
		if tdiff >= max_wait_sec:   #if count > max_tries:
			error_msg = "Timeout waiting to create snapshot."
			dbg_msg("ntnx_ss_create: %s" % error_msg)
			return (None, error_msg)

		# sleep the first iteration to give the server a chance to complete task
		if count < len(delay_list):
			wait_time = delay_list[count]
		else:
			wait_time = delay_list[-1]
		dbg_msg("delay %d sec" % wait_time)
		time.sleep(wait_time) # seconds

		(ss_info, error_msg) = ntnx_ss_get(ss_uuid)
		if (error_msg != None):
			return (None, error_msg)

		(ss_state, error_msg) = ntnx_ss_validate_is_complete(ss_info)
		if not (ss_state in ["kPending", "kRunning", "kComplete",
					"PENDING", "RUNNING", "COMPLETE"]
):
			error_msg = "Failed to create snapshot."
			dbg_msg("ntnx_ss_create: %s" % error_msg)
			return (None, error_msg)

		count += 1


	dbg_msg("Leaving ntnx_ss_create")
	return (ss_info, error_msg)

def ntnx_ss_validate_is_complete(ss_info):
	try:
		ss_state = ss_info['status']['state']
		#++ these are not available until kComplete
		#ss_name = ss_info['status']['name']
		#ss_type = ss_info['status']['snapshot_type']
		#--
		ss_name = ss_info['spec']['name']
		ss_type = ss_info['spec']['snapshot_type']
		ss_uuid = ss_info['metadata']['uuid']
		dbg_msg("ss_name:%s ss_state:%s ss_type:%s ss_uuid:%s" % (ss_name, ss_state, ss_type, ss_uuid))
	except:
		dbg_msg("Failed to get ss uuid")
		return None, "Failed to get ss uuid"

	if ss_state in ["kPending", "PENDING"]:
		dbg_msg("Snapshot is pending")

	elif ss_state in ["kRunning", "RUNNING"]:
		dbg_msg("Snapshot is running")

	elif ss_state not in ["kComplete", "COMPLETE"]:
		dbg_msg("Snapshot is not complete: %s" % ss_state)
		
	return ss_state, None

def ntnx_get_crt_info(ss_file_path, reference_ss_file_path, start_offset, end_offset=0):
	dbg_msg("Entering ntnx_get_crt_info")

	# Build the URL to get the appropriate information
	url = gen_url_ntnx_api_v3() + "/" + CONST_DATA_CHANGED_REGIONS

	if end_offset > 0:
		body = {
			"snapshot_file_path" : ss_file_path,
			"reference_snapshot_file_path" : reference_ss_file_path,
			"start_offset": start_offset,
			"end_offset" : end_offset
		}
	else:
		body = {
			"snapshot_file_path" : ss_file_path,
			"reference_snapshot_file_path" : reference_ss_file_path,
			"start_offset": start_offset,  # default 0
			# "end_offset" : 0    # default eof
		}

	(result, error_msg) = get_rest_api_response(url, body)
	if (error_msg != None):
		return (None, error_msg)

	return result, None

def ntnx_get_crt_info_thread(ss_file_path, reference_ss_file_path, start_offset, end_offset):
	dbg_msg("Entering ntnx_get_crt_info_thread: %s, %s, %d, %d" % (ss_file_path, reference_ss_file_path, start_offset, end_offset))

	error_msg = None
	lside = []
	crt_info_list = []

	next_offset = start_offset
	while next_offset != 0 and next_offset < end_offset:

		start_offset = next_offset

		(crt_info, error_msg) = ntnx_get_crt_info(ss_file_path, reference_ss_file_path, start_offset, end_offset)
		if (error_msg != None):
			dbg_msg("ntnx_get_crt_info_thread got error: %s" % (error_msg))
			break

		# Next_offset : This field specifies the offset from where the client
		# must continue the request. This field will not be set or will be set
		# to 0 in the following cases:
		# (a) When the file_size is less than 512 KB.
		# (b) When the complete file has been covered by the diff computation.
		if 'next_offset' in crt_info:
			next_offset = crt_info['next_offset']
		else:
			next_offset = 0
			
		file_size = crt_info['file_size']

		crt_info_list.append( crt_info )

	lside += [ crt_info_list, error_msg ]
	dbg_msg("Leaving ntnx_get_crt_info_thread: %s, %s, %d, %d" % (ss_file_path, reference_ss_file_path, start_offset, end_offset))
	return lside

def ntnx_get_crt_info_process(recv_p, send_p, ss_file_path, reference_ss_file_path, start_offset, end_offset):
	out = StringIO()
	sys.stdout = out
	err = StringIO()
	sys.stderr = err
	lside = None
	returnValue = 0

	try:
		recv_p.close()
		lside = ntnx_get_crt_info_thread(ss_file_path, reference_ss_file_path, start_offset, end_offset)
	finally:
		# Restore stdout, stderr and send their contents.
		sys.stdout = sys.__stdout__
		sys.stderr = sys.__stderr__

		outvalue = json.dumps(lside)
		result = [ outvalue, err.getvalue() ]
		try:
			send_p.send( json.dumps(result) )
		except Exception as e:
			dbg_msg("exception from pipe.send: %s" % (e))
			returnValue = -1
		send_p.close()
	return returnValue

def ntnx_get_crt_info_multiprocess(max_procs, ss_file_path, reference_ss_file_path, start_offset, end_offset, file_size):
	dbg_msg("Entering ntnx_get_crt_info_multiprocess: %s, %s, %d, %d" % (ss_file_path, reference_ss_file_path, start_offset, end_offset))

	crt_info_list = []
	one_gb = 1024*1024*1024

	if (max_procs < 1):
		max_procs = 1
	remaining = end_offset - start_offset
	look_size = remaining / max_procs
	look_size = ((look_size + (one_gb - 1)) / one_gb) * one_gb   # next 1 GB alignment
	if (look_size < (10 * one_gb)):
		look_size = 10 * one_gb

	### for testing
	### look_size = 10 * 1024 * 1024 * 1024
	### look_size = 512 * 1024 * 1024
	### look_size = 256 * 1024 * 1024

	dbg_msg("get crt max_procs: %d" % max_procs )
	dbg_msg("get crt look_size: %d" % look_size )


	while start_offset < file_size:

		pipe_procs_list = []

		# we need a try-except around the sub processes, so if there is an error in the parent,
		# then it will terminate the child processes before it calls sys.exit so that it doesn't deadlock.
		try:
			for i in range(max_procs):
				if start_offset >= file_size:
					break

				end_offset = start_offset + look_size
				if (end_offset >= file_size):
					end_offset = file_size

				recv_p, send_p = Pipe()

				dbg_msg("start process: %s, %s, %d, %d" % (ss_file_path, reference_ss_file_path, start_offset, end_offset))
				p = Process(target=ntnx_get_crt_info_process, args=(recv_p, send_p, ss_file_path, reference_ss_file_path, start_offset, end_offset))
				try:
					p.start()
				except OSError as e:
					dbg_msg("Couldn't start subprocess. got OSError exception: %s" % (e))
					if e.errno == errno.ENOMEM:
						# don't use any more processes just now
						recv_p.close()
						send_p.close()
						if i > 0:
							dbg_msg("set limit max_procs: %d" % (i))
							max_procs = i
							break
					error_msg = "Couldn't start subprocess: %s" % (e)
					error_msg = append_error_msg(error_msg, "Couldn't get changed regions.")
					return None, error_msg

				send_p.close()

				pipe_procs_list.append( [recv_p, p ] );

				start_offset += look_size

			error_msg = None
			for recv_p, proc in pipe_procs_list:

				dbg_msg("before get from pipe")
				result = None
				try:
					result = recv_p.recv()
				except EOFError:
					dbg_msg("EOF from pipe")
				###dbg_msg("RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR")
				###dbg_msg(result)
				###dbg_msg("RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR")
				if result == None:
					dbg_msg("got result=None")
					error_msg = "No result returned from process getting changed regions."
					break

				data = json.loads(result)
				out, err = data
				dbg_msg("after get from pipe")

				if err != None and len(err) > 0:
					# the stderr output is the logging output from subprocesses
					# add subprocess stderr to the main log
					dbg_msg( "stderr from subprocesses:\n%s" % (err) )

				if out != None and len(out) > 0:
					####dbg_msg("HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH")
					####dbg_msg(out)
					####dbg_msg("HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH")

					lside = json.loads(out)
					if lside == None or len(lside) < 2:
						dbg_msg("Failed to get result from out pipe")
						error_msg = "Couldn't get changed regions."
						break

					###dbg_msg("LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL")
					###dbg_msg("" + str(lside) )

					crt_info_list_2, error_msg = lside
					if error_msg:
						dbg_msg("a subprocess got an error: %s" % (error_msg))
						error_msg = append_error_msg(error_msg, "Couldn't get changed regions.")
						break

					crt_info_list += crt_info_list_2

				recv_p.close()
		except Exception as e:
			# put traceback in log
			ex_string = traceback.format_exc()
            		dbg_msg("Error:\n%s" % ex_string)
			# interpret exception 
			if isinstance(e, MemoryError):
				error_msg = "Out of memory"
			elif isinstance(e, OSError):
				error_msg = "Error: %s" % (e)
			else:
				error_msg = "Internal error."
			error_msg = append_error_msg(error_msg, "Couldn't get changed regions.")
			dbg_msg(error_msg)
			# terminate subprocesses
			for q,p in pipe_procs_list:
				p.terminate()
				p.join(10)
			# return error
			return None, error_msg

		# check for error
		if error_msg != None:
			for q,p in pipe_procs_list:
				p.terminate()
			return None, error_msg

		dbg_msg("wait for subprocesses")
		for q,p in pipe_procs_list:
			try:
				p.join(10)
			except TimeoutError:
				if error_msg == None:
					error_msg = "Timeout waiting for process."
			if p.exitcode != 0:
				dbg_msg("subprocess non-zero exitcode: %d" % p.exitcode)
				error_msg = "Couldn't get changed regions."
				return None, error_msg
		dbg_msg("all subprocesses complete")
		if error_msg != None:
			return None, error_msg

	return crt_info_list, None

def ntnx_ss_info_to_vg_disk_list(ss_info):

	# ss_info example:
	# ss_info value of None indicates to use simulation for development.
	if not ss_info:
	    ss_info = {
	    "status": {
	        "name": "pm-ubuntu-snapshot-01",
	        "state": "kComplete",
	        "snapshot_type": "CRASH_CONSISTENT",
	        "data_snapshot_identifier": "4f971ade-1e48-4742-857c-350c5dcdf847",
	        "snapshot_file_list": [
	            {
	                "file_path": "/NutanixManagementShare/.acropolis/vmdisk/036a5260-7b51-4545-9054-edbb88626699",
	                "snapshot_file_path": "/NutanixManagementShare/.snapshot/71/82835-1492071476401321-3171/.acropolis/vmdisk/036a5260-7b51-4545-9054-edbb88626699"
	            }
	        ],
	        "resources": {
	            "entity_uuid": "84991b70-866d-4ab2-aec8-06892e145809"
	        },
	        "expiration_time_msecs": 3652812494380
	    },
	    "spec": {
	        "resources": {
	            "entity_uuid": "84991b70-866d-4ab2-aec8-06892e145809"
	        },
	        "snapshot_type": "CRASH_CONSISTENT",
	        "name": "pm-ubuntu-snapshot-01"
	    },
	    "api_version": "3.0",
	    "metadata": {
	        "owner_reference": {
	            "kind": "user",
	            "uuid": "7e5d860a-11a0-4434-b14c-bdeba8d5605a",
	            "name": "admin"
	        },
	        "kind": "vm_snapshot",
	        "entity_version": 1,
	        "uuid": "a02e50f8-d169-4fc9-bd6a-9ddf7adb36f4",
	        "categories": {}
	    }
	}

	# get ss_uuid
	try:
		ss_uuid = ss_info['metadata']['uuid']
	except:
		dbg_msg("Failed to get ss uuid")
		return None, None, "Failed to get ss uuid"

	# volume group disk_list example:
	#"disk_list": [
	#	{
	#	"clone_config": {
	#		"disk_address": {
	#			"ndfs_filepath":
	#			"/NutanixManagementShare/.snapshot/71/82835-1492071476401321-3171/.acropolis/vmdisk/036a5260-7b51-4545-9054-edbb88626699"
	#		}
	#	},
	#	"index": 0
	#	}
	#],

	# build vg disk list from snapshot info
	disk_list = []
	try:
		ss_file_list = ss_info['status']['snapshot_file_list']
		count = 0
		for item in ss_file_list:
			disk_address_item = { "ndfs_filepath" : item['snapshot_file_path'] }
			clone_config = {
				"clone_config" : {
					"disk_address": {
						"ndfs_filepath" : item['snapshot_file_path']
					}
				}
			}
			disk_list.append( {
				"clone_config": { "disk_address": disk_address_item },
				"index": count
				} )
			count = count + 1
	except:
		dbg_msg("Failed to get disk list")
		return None, None, "Failed to get disk list"

	return ss_uuid, disk_list, None

def gen_iqn_prefix(vm_uuid):
	#return "prefix"
	return "vm-" + vm_uuid


def rest_api_with_task_completion(url, body, method=None):

	# rest request
	(result, error_msg) = get_rest_api_response(url, body, method)
	if (error_msg != None):
		return (None, error_msg)

	# track completion status
	(result, error_msg) = ntnx_track_completion_status(result)
	if not error_msg:
		dbg_msg("Completed")
	else:
		dbg_msg("Failed to complete")

	return result, error_msg

def rest_api_with_task_completion_with_retry(url, body, method=None):
	result = None
	progress_status = None

	retry_count = gTaskRetryCount
	if retry_count < 0:
		retry_count = 0
	max_tries = 1 + retry_count

	for count in range(0, max_tries):
		(result, error_msg) = rest_api_with_task_completion(url, body, method)
		if result != None:
			progress_status = result.get('progress_status')
		if progress_status and progress_status in ["Failed", "Aborted"]:
			if (count < max_tries):
				delaysec = 1 + count*3
				dbg_msg("Delay %d sec, then retry request" % delaysec)
				time.sleep( delaysec )  # seconds
		else:
			break
	return result, error_msg

def ntnx_create_volume_group(ss_info, vg_name, vg_uuid, client_ip_addr_list):
	dbg_msg("Entering ntnx_create_volume_group( %s, %s)" % (vg_uuid, vg_name))

	vm_uuid = options.uuid
	prefix = gen_iqn_prefix(vm_uuid)

	# Build the URL to get the appropriate information
	url = gen_url_ntnx_api_v2() + "/" + CONST_VOLUME_GROUPS

	# prep ss_uuid and disk_info
	(ss_uuid, disk_list, error_msg) = ntnx_ss_info_to_vg_disk_list(ss_info)
	if error_msg != None:
		return None, error_msg

	# prep attached_clients
	attached_clients = []
	for ipaddr in client_ip_addr_list:
		attached_clients.append( { "client_address": ipaddr } )

	# if client has multiple IP addresses, then set shared flag
	if len(attached_clients) > 1:
		is_shared = True
	else:
		is_shared = False

	body = {
		# add the attached_clients here to save a step, or else they could be added later with vg open
		"attached_clients": attached_clients,
		"description": "VG for backing up snapshot " + ss_uuid,
		"disk_list": disk_list,
		"flash_mode_enabled": False,
		"is_shared": is_shared,
		"iscsi_target_prefix": prefix,
		"name": vg_name,
		"uuid": vg_uuid
	}

	(result, error_msg) = rest_api_with_task_completion_with_retry(url, body)
	if (error_msg != None):
		return (None, error_msg)

	dbg_msg("Leaving ntnx_create_volume_group")
	return (1, error_msg)

def ntnx_vg_create(vm_uuid, description, disk_list, vg_name, vg_uuid, client_ip_addr_list):
	dbg_msg("Entering ntnx_vg_create( %s, %s)" % (vg_uuid, vg_name))

	#vm_uuid = options.uuid
	prefix = gen_iqn_prefix(vm_uuid)

	# Build the URL to get the appropriate information
	url = gen_url_ntnx_api_v2() + "/" + CONST_VOLUME_GROUPS

	# prep attached_clients
	attached_clients = []
	for ipaddr in client_ip_addr_list:
		attached_clients.append( { "client_address": ipaddr } )

	# if client has multiple IP addresses, then set shared flag
	if len(attached_clients) > 1:
		is_shared = True
	else:
		is_shared = False

	body = {
		# add the attached_clients here to save a step, or else they could be added later with vg open
		"attached_clients": attached_clients,
		"description": description,
		"disk_list": disk_list,
		"flash_mode_enabled": False,
		"is_shared": is_shared,
		"iscsi_target_prefix": prefix,
		"name": vg_name,
		"uuid": vg_uuid
	}

	(result, error_msg) = rest_api_with_task_completion_with_retry(url, body)
	if (error_msg != None):
		return (None, error_msg)

	dbg_msg("Leaving ntnx_vg_create")
	return (1, error_msg)

def ntnx_get_volume_group(vg_uuid):
	dbg_msg("Entering ntnx_get_volume_group( %s)" % vg_uuid)

	url = gen_url_ntnx_api_v2() + "/" + CONST_VOLUME_GROUPS + "/" + vg_uuid

	(result, error_msg) = get_rest_api_response(url)
	if (error_msg != None):
		return (None, error_msg)

	dbg_msg("Leaving ntnx_get_volume_group")
	return (result, error_msg)

def ntnx_get_volume_groups():
	url = gen_url_ntnx_api_v2() + "/" + CONST_VOLUME_GROUPS
	(result, error_msg) = get_rest_api_response(url)
	if (error_msg != None):
		return (None, error_msg)
	return (result, error_msg)

def ntnx_modify_volume_group(body, vg_name, vg_uuid):
	dbg_msg("Entering ntnx_modify_volume_group( %s, %s)" % (vg_uuid, vg_name))

	# Build the URL to get the appropriate information
	url = gen_url_ntnx_api_v2() + "/" + CONST_VOLUME_GROUPS + "/" + vg_uuid

# {
#     "iscsi_target": "prefix-a272600d-10f4-43da-ab36-f3ba733f421d",
#     "logical_timestamp": 1,
#     "description": "VG for backing up snapshot a02e50f8-d169-4fc9-bd6a-9ddf7adb36f4",
#     "uuid": "a272600d-10f4-43da-ab36-f3ba733f421d",
#     "name": "pm-test-vg",
#     "flash_mode_enabled": false,
#     "attachment_list": [
#         {
#             "initiator_ip_address": "192.168.134.116",
#             "client_uuid": "6773f11d-03dc-4b4e-84ad-8f0beca9ec7d"
#         }
#     ],
#     "is_shared": false,
#     "disk_list": [
#         {
#             "index": 0,
#             "vmdisk_uuid": "20869e03-7d01-4533-9660-c043069a1383",
#             "container_uuid": "63feb612-942c-443a-89fe-b5fcf39db5c5",
#             "container_id": 5696,
#             "storage_container_id": 5696,
#             "vmdisk_size_mb": 20480,
#             "flash_mode_enabled": false,
#             "storage_container_uuid": "63feb612-942c-443a-89fe-b5fcf39db5c5"
#         }
#     ]
# }
			#entity = json.dumps(result.get('entity_list')[-1])
			#entity_list = json.loads(entity)
			#return entity_list.get('entity_id')

	body['attachment_list'] = []
	# TODO: deep copy?
	del body["iscsi_target"]  # can't modify iscsi_target

	#item = result.get('attachment_list')[0]
	#item.

#	body = {
#		"attachment_list": [
#			#{
#			#"client_address": ""
#			#}
#		] # ,
#		#"description": "VG for backing up snapshot a02e50f8-d169-4fc9-bd6a-9ddf7adb36f4",
#		#"disk_list": [
#		#	{
#		#	"clone_config": {
#		#	"disk_address": {
#		#		"ndfs_filepath":
#		#		"/NutanixManagementShare/.snapshot/71/82835-1492071476401321-3171/.acropolis/vmdisk/036a5260-7b51-4545-9054-edbb88626699"
#		#		}
#		#	},
#		#	"index": 0
#		#}
#		#],
#		#"flash_mode_enabled": False,
#		#"is_shared": False,
#		#"iscsi_target_prefix": "prefix",
#		#"name": vg_name,
#		#"uuid": vg_uuid
#	}

	(result, error_msg) = rest_api_with_task_completion_with_retry(url, body)
	if (error_msg != None):
		return (1, error_msg)

	dbg_msg("Leaving ntnx_modify_volume_group")
	return (1, error_msg)

def ntnx_close_volume_group(result, vg_name, vg_uuid):
	dbg_msg("Entering ntnx_close_volume_group( %s, %s)" % (vg_uuid, vg_name))

	# Build the URL to get the appropriate information
	url = gen_url_ntnx_api_v2() + "/" + CONST_VOLUME_GROUPS + "/" + vg_uuid + "/" + "close"
	#dbg_msg("URL for request: %s" % url)

# {
#     "iscsi_target": "prefix-a272600d-10f4-43da-ab36-f3ba733f421d",
#     "logical_timestamp": 1,
#     "description": "VG for backing up snapshot a02e50f8-d169-4fc9-bd6a-9ddf7adb36f4",
#     "uuid": "a272600d-10f4-43da-ab36-f3ba733f421d",
#     "name": "pm-test-vg",
#     "flash_mode_enabled": false,
#     "attachment_list": [
#         {
#             "initiator_ip_address": "192.168.134.116",
#             "client_uuid": "6773f11d-03dc-4b4e-84ad-8f0beca9ec7d"
#         }
#     ],
#     "is_shared": false,
#     "disk_list": [
#         {
#             "index": 0,
#             "vmdisk_uuid": "20869e03-7d01-4533-9660-c043069a1383",
#             "container_uuid": "63feb612-942c-443a-89fe-b5fcf39db5c5",
#             "container_id": 5696,
#             "storage_container_id": 5696,
#             "vmdisk_size_mb": 20480,
#             "flash_mode_enabled": false,
#             "storage_container_uuid": "63feb612-942c-443a-89fe-b5fcf39db5c5"
#         }
#     ]
# }

	error_msg = None
	attachment_list = []
	try:
		attachment_list = result["attachment_list"]
		client_ip_addr = result["attachment_list"][0]["initiator_ip_address"]
	except:
		dbg_msg("Failed to get initiator_ip_address from vg info")
		return 0, None

	if not client_ip_addr:
		return 0, None

	for attachment in attachment_list:
		client_ip_addr = attachment["initiator_ip_address"]
		dbg_msg("vg_close for client_ip_addr: %s" % client_ip_addr)

		body = {
			"iscsi_client": {
				"client_address": client_ip_addr
			},
			"iscsi_client_ip": client_ip_addr,
			"logical_timestamp": 0,
			"operation": "ATTACH",
			"uuid": vg_uuid
		}
		
		(result, error_msg) = get_rest_api_response(url, body)
		if (error_msg != None):
			return (1, error_msg)
		else:
			dbg_msg("Successfully volume_group closed: %s" % client_ip_addr)

	dbg_msg("Leaving ntnx_close_volume_group")
	return (0, error_msg)

def ntnx_delete_volume_group(vg_name, vg_uuid):
	dbg_msg("Entering ntnx_delete_volume_group( %s, %s)" % (vg_uuid, vg_name))

	# Build the URL to get the appropriate information
	url = gen_url_ntnx_api_v2() + "/" + CONST_VOLUME_GROUPS + "/" + vg_uuid

	(result, error_msg) = rest_api_with_task_completion_with_retry(url, None, "DELETE")
	if (error_msg != None):
		return (1, error_msg)

	dbg_msg("Leaving ntnx_delete_volume_group")
	return (1, error_msg)

def ntnx_vm_get(vm_uuid):
	if vm_uuid != None:
		url = gen_url_ntnx_api_v2() + "/" + CONST_VMS + "/" + vm_uuid
	else:
		url = gen_url_ntnx_api_v2() + "/" + CONST_VMS
	(vm_info, error_msg) = get_rest_api_response(url)
	if (error_msg != None):
		return None, error_msg
	return vm_info, None

def ntnx_vm_get_info(vm_uuid):

	url = gen_url_ntnx_api_v2() + "/" + CONST_VMS + "/" + vm_uuid + "/?" + CONST_GET_VM_DISKS
	(vm_info, error_msg) = get_rest_api_response(url)
	if (error_msg != None):
		return None, error_msg

	return vm_info, None


def ntnx_vm_create(vm_info):
	dbg_msg("Entering ntnx_vm_create")

	# Build the URL to get the appropriate information
	url = gen_url_ntnx_api_v2() + "/" + CONST_VMS
	body = vm_info

	(result, error_msg) = rest_api_with_task_completion_with_retry(url, body)
	if (error_msg != None):
		return (None, error_msg)

	if result != None:
		entity = result.get('entity_list')[0]
		entity_id = entity.get('entity_id')

	dbg_msg("Leaving ntnx_vm_create")
	return (entity_id, error_msg)

def ntnx_vm_attach_disks(vm_uuid, vm_disks):
	url = gen_url_ntnx_api_v2() + "/" + CONST_VMS + "/" + vm_uuid + "/" + CONST_VMS_DISKS_ATTACH
	body = {
		"uuid" : vm_uuid,
		"vm_disks": vm_disks
	}
	(result, error_msg) = rest_api_with_task_completion_with_retry(url, body)
	if (error_msg != None):
		return (None, error_msg)
	return result, error_msg

def ntnx_vm_update_disk_info(vm_uuid, vm_disk_info):
	url = gen_url_ntnx_api_v2() + "/" + CONST_VMS + "/" + vm_uuid + "/"
	body = {
		"vm_disk_info" : vm_disk_info,
	}
	(result, error_msg) = rest_api_with_task_completion_with_retry(url, body, "PUT")
	if (error_msg != None):
		return (None, error_msg)
	return result, error_msg

def ntnx_vm_update_boot(vm_uuid, bootvals):
	url = gen_url_ntnx_api_v2() + "/" + CONST_VMS + "/" + vm_uuid + "/"
	body = {
		"boot" : bootvals,
	}
	(result, error_msg) = rest_api_with_task_completion_with_retry(url, body, "PUT")
	if (error_msg != None):
		return (None, error_msg)
	return result, error_msg


	#p = subprocess.Popen('ls', shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
	##list = p.stdout.readlines()
	##list = p.read().splitlines()
	#list = commands.getoutput('ls').splitlines()
	#print list
	##for line in p.stdout.readlines():
	##	print line,
	#retval = p.wait()

	# from subprocess import Popen, PIPE
	# cmd = "ls -l ~/"
	# p = Popen(cmd , shell=True, stdout=PIPE, stderr=PIPE)
	# out, err = p.communicate()
	# print "Return code: ", p.returncode
	# print out.rstrip(), err.rstrip()

def do_cmd(cmd):
	lines = None
	error_msg = None

	dbg_msg("Run command: '%s'" % cmd)
	p = Popen(cmd , shell=True, stdout=PIPE, stderr=PIPE)
	out, err = p.communicate()
	dbg_msg( "Return code: %d" % p.returncode )
	dbg_msg( "stdout: '%s'" % out.rstrip() )
	dbg_msg( "stderr: '%s'" % err.rstrip() )
	if 0 != p.returncode:
		error_msg = out
		if (error_msg and len(error_msg) > 0 and error_msg[-1] != '\n'):
			error_msg += '\n'
		error_msg += err.rstrip()
		return None, error_msg

	lines = out.splitlines()
	return lines, error_msg

def do_cmd_expect_output(cmd, hideStdout=False):
	lines = None
	error_msg = None

	dbg_msg("Run command: '%s'" % cmd)
	p = Popen(cmd , shell=True, stdout=PIPE, stderr=PIPE)
	out, err = p.communicate()
	dbg_msg( "Return code: %d" % p.returncode )
	if hideStdout:
		dbg_msg( "stdout: (hidden)" )
	else:
		dbg_msg( "stdout: '%s'" % out.rstrip() )
	dbg_msg( "stderr: '%s'" % err.rstrip() )
	if 0 != p.returncode or len(out) == 0:
		error_msg = out
		if (error_msg and len(error_msg) > 0 and error_msg[-1] != '\n'):
			error_msg += '\n'
		error_msg += err.rstrip()
		return None, error_msg

	lines = out.splitlines()
	return lines, error_msg

def do_iscsi_cmd_expect_output_with_retry(cmd, hideStdout=False):
	lines = None
	error_msg = None

	retry_count = gIscsiCmdRetryCount
	if retry_count < 0:
		retry_count = 0
	max_tries = 1 + retry_count

	for count in range(0, max_tries):
		lines, error_msg = do_cmd_expect_output(cmd)
		if (not lines) or (len(lines) == 0) or (error_msg != None):
			if (count < max_tries):
				delaysec = 1 + count*3
				dbg_msg("Delay %d sec, then retry request" % delaysec)
				time.sleep( delaysec )  # seconds
		else:
			break

	return lines, error_msg

def ini_get_int(section, key, default):
	value = default
	# get snapshot timeout value from master.ini
	cmd = "/usr/bp/bin/bputil -g %s %s %d /usr/bp/bpinit/master.ini" % (section, key, default)
	lines, error_msg = do_cmd_expect_output(cmd)
	if (not lines) or (len(lines) == 0) or (error_msg != None):
		error_msg = "Couldn't get %s from master.ini" % key
		dbg_msg(error_msg)
	else:
		value = int( lines[0] )
	return value

def get_xml_from_iscsi_disk_devices(iscsi_disk_devices, diskindexmap=None):
	idd_xml = Element('iscsi_disk_devices')

	size = len(iscsi_disk_devices)
	for index in range(0, size):
		disk_device = iscsi_disk_devices[index]
		item_xml = SubElement(idd_xml, 'iscsi_disk_device')
		if diskindexmap != None:
			key = diskindexmap[index]['key']
		else:
			key = index
		SubElementWithText(item_xml, 'key', str(key))
		SubElementWithText(item_xml, 'path', disk_device)

	return idd_xml

# Change dir to empty dir
# as workaround to --sid interpretation by iscsiadm.
# The iscsiadm command treats the --sid value as either a path or a number.
# This overloaded interpretation is unfortunate, because 
# if "/" is the current working directory, that directory sometimes 
# has a file named "1" in it as a result of a typo in some script somewhere.
# This causes iscsiadm to return an error for the very first sid
# where the error is "iscsiadm: 1 is not a directory".
# As a workaround, chdir to another directory that is less likely to have that file in it.
def chdir_to_empty():
	empty_dir = '/var/run/aprotect/empty'
	if not os.path.exists(empty_dir):
		try:
    			os.makedirs(empty_dir)
		except:
			dbg_msg("Couldn't make dir %s" % empty_dir)
	# change the directory
	try:
		os.chdir( empty_dir )
	except:
		dbg_msg("Couldn't chdir to %s" % empty_dir)

	# get current working directory.
	retval = os.getcwd()
	dbg_msg( "Current working directory: '%s'" % retval )

# target iqn sort function
# example: ...-f2cdf0e854ac-tgt12 
#    tgt0, tgt1, tgt10, tgt11, tgt2  --> sort --> tgt0, tgt1, tgt2, tgt10, tgt11
def myTargetIqnSort(e):
	ix = e.rfind('-')
	if ix >=0:
		tgt = e[ix+1:]  		# get tgt string, i.e. tgt12
		if tgt.startswith("tgt"):
			val = int( tgt[3:] )	# get number, i.e. 12
			return val
	return e

def iscsi_attach(iscsi_ip_addr, vg_uuid):
	# input
	# iscsi_ip_addr = server_data_ip_addr # "192.168.171.241"
	# iscsi_ip_addr = "192.168.171.244"
	iscsi_port_str = "3260"

	storage_name = "ntnx-server-%s" % iscsi_ip_addr.replace('.', '-')

	chdir_to_empty()

	# find target IQNs
	# cmc_iscsi discover ip port
	cmd = "cmc_iscsi discover " + iscsi_ip_addr + " " + iscsi_port_str
	lines, error_msg = do_iscsi_cmd_expect_output_with_retry(cmd)
	if (not lines) or (len(lines) == 0) or (error_msg != None):
		error_msg += "\nCouldn't discover iscsi target"
		return 1, error_msg

	dbg_msg("iscsi targets: %s" % lines)
	iscsi_target_iqns = lines

	dbg_msg("vg_uuid: %s" % vg_uuid)

	# filter out target iqn's that do not belong to volume_group
	filtered_iqns = []
	for iscsi_target_iqn in iscsi_target_iqns:
		# check if target iqn belongs with the volume_group
		if vg_uuid != None:
			substr = vg_uuid + '-tgt'
			if not substr in iscsi_target_iqn:
				dbg_msg("Skipping iscsi target that is not part of volume group: %s" % iscsi_target_iqn)
				continue
		filtered_iqns.append( iscsi_target_iqn )

	# sort the iqns in target number order
	filtered_iqns.sort( key=myTargetIqnSort )

	iscsi_target_iqns = filtered_iqns

	iscsi_disk_devices = []

	for iscsi_target_iqn in iscsi_target_iqns:

		# list luns
		cmd = "cmc_iscsi listluns %s %s %s"  % (iscsi_target_iqn, iscsi_ip_addr, iscsi_port_str)
		lines, error_msg = do_iscsi_cmd_expect_output_with_retry(cmd)
		if (not lines) or (len(lines) == 0) or (error_msg != None):
			error_msg += "\nCouldn't list luns"
			return 1, error_msg
		iscsi_luns = lines
		dbg_msg("iscsi_luns: %s" % iscsi_luns)

		# connect to luns
		for lun in iscsi_luns:
			# The 8th param value of 2 means show the device path.
			cmd = "cmc_iscsi connect %s source %s %s %s %s 2" % \
				(storage_name, iscsi_ip_addr, iscsi_port_str, iscsi_target_iqn, lun)
			lines, error_msg = do_iscsi_cmd_expect_output_with_retry(cmd)
			if (not lines) or (len(lines) == 0) or (error_msg != None):
				error_msg += "\nCouldn't connect to iscsi target"
				return 1, error_msg

			# last line will have the device path
			disk = lines[-1].strip()
			if not "/dev/" in disk:
				return 1, "Couldn't get disk path for lun"
			dbg_msg("Got lun %s disk %s" % (lun, disk) )
			ilun = int(lun)
			iscsi_disk_devices.append( disk )

	# show results
	dbg_msg("iscsi disk devices:")
	for ix in range(0, len(iscsi_disk_devices)):
		disk = iscsi_disk_devices[ix]
		dbg_msg("  index %s disk %s" % (ix, disk) )

	return iscsi_disk_devices, error_msg

def iscsi_get_session_targets(iscsi_ip_addr, vm_uuid):
	# use "iscsiamd -m session -P0" to discover attached targets
	# filter target iqns using vg_uuid
	# filter target iqns using prefix
	# example line:
	#   tcp: [198] 192.168.171.241:3260,1 iqn.2010-06.com.nutanix:prefix-ab4ed537-fed8-4895-a338-04693a872a40-tgt0 (non-flash)

	cmd = "iscsiadm -m session -P0"
	lines, error_msg = do_cmd(cmd)
	if (not lines) or (error_msg != None):
		if error_msg != None and 'No active sessions' in error_msg:
			return [], None
		error_msg = append_error_msg(error_msg, "Couldn't get iscsi sessions")
		return None, error_msg

	# gen prefix based on vm_uuid
	prefix = gen_iqn_prefix(vm_uuid)

	target_iqn_list = []

	for line in lines:
		try:
			line_parts = line.split()
			ipaddr = line_parts[2].split(':')[0]
			# filter ip address
			if not iscsi_ip_addr == ipaddr:
				continue
			# filter iqn prefix
			iqn = line_parts[3]
			if not prefix in iqn:
				continue
		except:
			dbg_msg("iscsi_get_sessions: caught exception while parsing line '%s'" % line)
			continue
		# found match, add it to list
		dbg_msg("iscsi_get_sessions: found match '%s'" % iqn)
		target_iqn_list.append(iqn)

	return target_iqn_list, None

def iscsi_detach(iscsi_ip_addr, vm_uuid):
	# input
	iscsi_port_str = "3260"

	chdir_to_empty()

	if 1:
		# get target iqns from active sessions
		iscsi_target_iqns, error_msg = iscsi_get_session_targets(iscsi_ip_addr, vm_uuid)
		if error_msg != None:
			error_msg += "\nCouldn't get iscsi session targets"
			return 1, error_msg
	else:
		# cmc_iscsi discover ip port
		cmd = "cmc_iscsi discover " + iscsi_ip_addr + " " + iscsi_port_str
		lines, error_msg = do_iscsi_cmd_expect_output_with_retry(cmd)
		if (not lines) or (len(lines) == 0) or (error_msg != None):
			error_msg += "\nCouldn't discover iscsi target"
			return 1, error_msg
		iscsi_target_iqns = lines

	dbg_msg("iscsi targets: %s" % iscsi_target_iqns)

	# no need for disconnect for "source" usage of cmc_iscsi
	# cmc_iscsi disconnect pm-ntnx-test 

	# cmc_iscsi logout iqn.2010-06.com.nutanix:prefix-a272600d-10f4-43da-ab36-f3ba733f421d-tgt0 
	sum_error_msg = ""
	if iscsi_target_iqns != None:
		for iqn in iscsi_target_iqns:
			cmd = "cmc_iscsi logout %s" % iqn
			lines, error_msg = do_cmd(cmd)
			if (error_msg != None):
				error_msg += "\nCouldn't logout iscsi target %s" % iqn
				sum_error_msg += error_msg
			
	if len(sum_error_msg) > 0:
		return 1, sum_error_msg

	return 0, None

def get_adapter_ip_addresses( ip_addrs ):
	# get additional ip addresses
	# use:
	#   cmc_network adapter list
	#   cmc_network adapter info eth0
	error_msg = None
	cmd = "cmc_network adapter list"
	lines, error_msg = do_cmd_expect_output(cmd)
	if (not lines) or (len(lines) == 0) or (error_msg != None):
		error_msg = "Couldn't get network adapter list"
		dbg_msg( error_msg )
		return ip_addrs, error_msg

	for adapter in lines:
		cmd = "cmc_network adapter info %s" % adapter
		lines, error_msg = do_cmd_expect_output(cmd)
		if lines == None or len(lines) == 0 or (error_msg != None):
			if error_msg != None:
				error_msg += "\n"
			else:
				error_msg = ""
			error_msg += "Couldn't get info for network adapter %s" % adapter
			dbg_msg( error_msg )
		else:
			for line in lines:
				parts = line.split()
				if (len(parts) > 1) and (parts[0] == 'ipaddr'):
					if parts[1] != 'unknown' and parts[1] not in ip_addrs:
						ip_addrs.append( parts[1] )

	return ip_addrs, error_msg


def get_dpu_ip_addresses():
	name = None
	longname = None
	ip_addrs = []
	names = {}

	# get hostname
	cmd = "cmc_network hostname info"
	dbg_msg("will run command: '%s'" % cmd)
	lines = commands.getoutput(cmd).splitlines()

	for ix in range(0,len(lines)):
		parts = lines[ix].split()
		if len(parts) < 2:
			continue
		if len(parts) == 2:
			names[ parts[0] ] = parts[1]
		if parts[0] == 'name':
			name = parts[1]
		if parts[0] == 'longname':
			longname = parts[1]

	if name == None:
		name = longname

	if name == None:
		cmd = "hostname"
		lines = commands.getoutput(cmd).splitlines()
		if lines == None or 0 == len(lines):
			error_msg = "Couldn't get hostname"
			return None, error_msg
		name = lines[0]

	# get ip addr
	cmd = "cmc_network host info %s" % name
	dbg_msg("will run command: '%s'" % cmd)
	lines = commands.getoutput(cmd).splitlines()

	if lines == None or 0 == len(lines):
		error_msg = "Couldn't get ip address for %s" % name
		return None, error_msg
	for ix in range(0,len(lines)):
		parts = lines[ix].split()
		if (len(parts) > 1) and (parts[0] == 'ip'):
			ip_addrs.append( parts[1] )

	# get additional ip addresses
	ip_addrs, error_msg = get_adapter_ip_addresses( ip_addrs )

	# if error_msg, then keep going

	# for test: ADD FAKE IP ADDRs
	# ip_addrs = [ ip_addrs[0], '192.168.134.115', '192.168.134.117' ]

	# return list
	return ip_addrs, None

def gen_id_name( ip_addr, vm_uuid, curr_time, type_string ):
	curr_time = "0"  # force constant result for now 
	id = "ub:%s:%s:%s:%s" % (ip_addr, vm_uuid, curr_time, type_string)
	return id

def doprmtest():
	dbg_msg("Executing prmtest")

	if options.prmtest in [ "all", "vg.up", "vg.down" ]:
		(vg_uuid, error_msg) = ntnx_create_idempotence_id("pm_test")
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

	vg_name = "pm-test-vg"
	client_ip_addr_list = [ "192.168.134.116" ]

	ss_info = None

	if options.prmtest in ["all", "vm.info"]:
		## get vm info
		(vm_info, vm_info_xml, error_msg) = get_xml_for_vm_info_for_metadata()
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

		## get vm hardware hash
		(hw_hash, error_msg) = gen_hardware_hash_string(vm_info)
		if error_msg != None:
			output_xml_for_error(error_msg)
			sys.exit(2)

	if options.prmtest in ["all", "dpu.info"]:
		(ip_addrs, error_msg) = get_dpu_ip_addresses()
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)
		dbg_msg("dpu ip_addrs: %s" % (ip_addrs,) )

	if options.prmtest in ["all", "ss.create", "crt.get"]:
		dbg_msg("ss.create")

		vm_uuid = "84991b70-866d-4ab2-aec8-06892e145809"

		ss_name = "pm-ss-test"
		(ss_uuid, error_msg) = ntnx_create_idempotence_id(ss_name)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

		(result, error_msg) = ntnx_ss_delete(ss_uuid)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

		(result, error_msg) = ntnx_ss_create(vm_uuid, ss_name, ss_uuid)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

	if options.prmtest in ["all", "ss.get", "crt.get"]:

		ss_uuid_1 = "a02e50f8-d169-4fc9-bd6a-9ddf7adb36f4"
		ss_uuid_2 = ss_uuid

		# get reference ss info
		(ss_result_1, error_msg) = ntnx_ss_get(ss_uuid_1)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)
		ss_info_1 = ss_result_1

		(ss_state_1, error_msg) = ntnx_ss_validate_is_complete(ss_info_1)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

		# get current ss info
		(result, error_msg) = ntnx_ss_get(ss_uuid)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)
		ss_info = result

		(result, error_msg) = ntnx_ss_validate_is_complete(ss_info)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

	if options.prmtest in ["all", "crt.get"]:

		dbg_msg("crt.get")
		ss_info_2 = ss_info

		(ss_uuid_1, disk_list_1, error_msg) = ntnx_ss_info_to_vg_disk_list(ss_info_1)
		if error_msg != None:
			return 1, error_msg

		(ss_uuid_2, disk_list_2, error_msg) = ntnx_ss_info_to_vg_disk_list(ss_info_2)
		if error_msg != None:
			return 1, error_msg

		# sample disk_list:
		#"disk_list": [
		#	{
		#	"clone_config": {
		#		"disk_address": {
		#			"ndfs_filepath":
		#			"/NutanixManagementShare/.snapshot/71/82835-1492071476401321-3171/.acropolis/vmdisk/036a5260-7b51-4545-9054-edbb88626699"
		#		}
		#	},
		#	"index": 0
		#	}
			
		# for each disk, get crt info
		nbr = len(disk_list_1)
		for index in range(0, nbr):
			if (index != disk_list_1[index]["index"]):
				dbg_msg("disk_list_1 index does not match expectation:  index %d != %d " % \
					index, disk_list_1[index]["index"] )
				sys.exit(2)

			reference_ss_file_path = disk_list_1[index]["clone_config"]["disk_address"]["ndfs_filepath"]
			ss_file_path = disk_list_2[index]["clone_config"]["disk_address"]["ndfs_filepath"]

			start_offset = 0
			(result, error_msg) = ntnx_get_crt_info(ss_file_path, reference_ss_file_path, start_offset)
			if (error_msg != None):
				output_xml_for_error(error_msg)
				sys.exit(2)

	if options.prmtest in ["all", "vg.up"]:

		(result, error_msg) = ntnx_create_volume_group(ss_info, vg_name, vg_uuid, client_ip_addr_list)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

	if options.prmtest in ["all", "iscsi.up", "iscsi.down"]:
		(target_ip_addr, error_msg) = ntnx_get_cluster_data_ip_addr()
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

	if options.prmtest in ["all", "iscsi.up"]:

		dbg_msg("iscsi.up")
		(result, error_msg) = iscsi_attach(target_ip_addr, vg_uuid)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

	if options.prmtest in ["all", "iscsi.down"]:
		dbg_msg("iscsi.down")
		(result, error_msg) = iscsi_detach(target_ip_addr, vm_uuid)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

		
	if options.prmtest in ["all", "vg.down"]:

		(result, error_msg) = ntnx_get_volume_group(vg_uuid)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

		(result, error_msg) = ntnx_close_volume_group(result, "vg_name", vg_uuid)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

		(result, error_msg) = ntnx_delete_volume_group(vg_name, vg_uuid)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

		if 0:
			(result, error_msg) = ntnx_delete_idempotence_id("pm_test")
			if (error_msg != None):
				output_xml_for_error(error_msg)
				sys.exit(2)

	if options.prmtest in [ 'add_teardown_entry' ]:
		entry = "this is a test:123"
		add_teardown_entry( get_teardown_id() + "|" + entry )

	if options.prmtest in [ 'remove_teardown_entry' ]:
		entry = "this is a test:123"
		remove_teardown_entry( get_teardown_id() + "|" + entry )

	if options.prmtest in [ 'process_idle_info_files' ]:
		process_idle_info_files() 

	if options.prmtest in [ 'credentials.lookup']:
		result, error_msg = get_server_credentials(options.server)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)
		dbg_msg("got: %s" % (result) )

	if options.prmtest in ["sleep"]:
		time.sleep( 10 )  # seconds
		time.sleep( 10 )  # seconds


	dbg_msg("Successful prmtest" )

def find_vm_disk_by_uuid( vm_disk_info, lookfor_uuid ):

	number_of_disks = len(vm_disk_info)
	for j in range(0,number_of_disks):
		try:
			vm_disk = vm_disk_info[j]

			# a cdrom without a disk will not have a vmdisk_uuid
			vmdisk_uuid = vm_disk['disk_address']['vmdisk_uuid']

			if (vmdisk_uuid == lookfor_uuid):
				return vm_disk

			#disk_size = vm_disk_info[j]['size']
			#disk_device_index = vm_disk_info[j]['disk_address']['device_index']
			#disk_label = vm_disk_info[j]['disk_address']['disk_label']
			#is_empty = vm_disk_info[j]['is_empty']

		except:
			dbg_msg("One or more parameters unavailable for disk %s" % j)
			continue

	return None


def get_xml_for_disks( vm_info, disk_list_1, disk_list_2 ):

	error_msg = None
	disks_backup_size = 0
	nbr_disks_included = 0
	is_full = True
	disks_xml = None

	# disks xml
	disks_xml = Element('disks')

	# full or incremental
	if disk_list_1:
		is_full = False 

	if not 'vm_disk_info' in vm_info:
		# return empty disks element
		return nbr_disks_included, disks_backup_size, disks_xml, error_msg

	# get cdrom exclude list
	vm_disk_info = vm_info['vm_disk_info']
	cdrom_uuid_list = get_vm_disk_cdrom_uuid_list( vm_disk_info )
	dbg_msg("cdrom_uuid_list: %s" % cdrom_uuid_list)

	# prepare exclude list
	exclude_list = []
	if options.exclude != None and len(options.exclude) > 0 and options.exclude != 'empty':
		exclude_list = options.exclude.split(':')

	# sample disk_list:
	#"disk_list": [
	#	{
	#	"clone_config": {
	#		"disk_address": {
	#			"ndfs_filepath":
	#			"/NutanixManagementShare/.snapshot/71/82835-1492071476401321-3171/.acropolis/vmdisk/036a5260-7b51-4545-9054-edbb88626699"
	#		}
	#	},
	#	"index": 0
	#	}

	naming_index = 0

	# for each disk, get crt info
	nbr = len(disk_list_2)
	for index in range(0, nbr):
		if (index != disk_list_2[index]["index"]):
			dbg_msg("disk_list_2 index does not match expectation:  index %d != %d " % \
				index, disk_list_2[index]["index"] )
			error_msg = "Internal error"
			return 0, 0, None, error_msg

		ss_file_path = disk_list_2[index]["clone_config"]["disk_address"]["ndfs_filepath"]

		# exctract uuid from ss_file_path.  the last 36 characters.
		ss_file_uuid = ss_file_path[-36:]

		dbg_msg("ss_file_uuid: %s" % ss_file_uuid)

		# locate vm_disk
		vm_disk = find_vm_disk_by_uuid( vm_disk_info, ss_file_uuid )
		if vm_disk != None:
			disk_label = vm_disk['disk_address']['disk_label']
		else:
			disk_label = 'none'

		# is uuid a cdrom
		if ss_file_uuid in cdrom_uuid_list:
			dbg_msg("Excluding cdrom disk: %s, %s" % (disk_label, ss_file_uuid))
			naming_index += 1
			continue

		# the ss_file_path contains disk uuid.
		dbg_msg("exclude_list: %s" % exclude_list)
		is_excluded = False
		for exclude_item in exclude_list:
			dbg_msg("exclude_item: %s" % exclude_item)
			dbg_msg("ss_file_path: %s" % ss_file_path)
			if ss_file_path.endswith(exclude_item):
				dbg_msg("Excluding disk: %s" % exclude_item)
				is_excluded = True

		if is_excluded:
			# naming index does not include cdrom
			naming_index += 1
			continue

		nbr_disks_included += 1

		# get uuid from ss_file_path
		uuid_from_ss_file_path = ""
		slash_index = ss_file_path.rfind('/')
		if (slash_index != -1) and ((slash_index+1) < len(ss_file_path)):
			uuid_from_ss_file_path = ss_file_path[slash_index+1:]
			
		# crt handling for full vs incr.  No reference snapshot for full backup.
		if is_full:
			# full backup
			reference_ss_file_path = "" #None
		else:
			# incremental backup
			# locate corresponding ndfs_filepath from previous snapshot.
			# an add or removal of cdrom affects the index location of the disks following it.
			reference_ss_file_path = None
			filepath = None
			for ref_disk_item in disk_list_1:
				filepath = ref_disk_item["clone_config"]["disk_address"]["ndfs_filepath"]
				slash_index = filepath.rfind('/')
				if (slash_index != -1) and ((slash_index+1) < len(filepath)):
					uuid_from_ref_file_path = filepath[slash_index+1:]
					if (uuid_from_ref_file_path == uuid_from_ss_file_path):
						# match
						reference_ss_file_path = filepath
						break
						
			# reference_ss_file_path = disk_list_1[index]["clone_config"]["disk_address"]["ndfs_filepath"]
			if reference_ss_file_path == None:
				# couldn't find reference in previous snapshot
				dbg_msg("Couldn't disk find file reference in previous snapshot")
				error_msg = "Couldn't find disk file reference in previous snapshot"
				return 0, 0, None, error_msg

		use_alignment = True
		is_skip_zeros = is_full
		crt_info_list = []
		crt_info_xml = None
		next_offset = 0 
		start_offset = -1

		use_multiprocess = False
		max_procs = get_multiprocess_max_procs()
		if (max_procs > 1):
			use_multiprocess = True

		if use_multiprocess:
			# get the first one. this will give us the file size.
			start_offset = next_offset
			(crt_info, error_msg) = ntnx_get_crt_info(ss_file_path, reference_ss_file_path, start_offset)
			if (error_msg != None):
				return 0, 0, None, error_msg

			# Next_offset : This field specifies the offset from where the client
			# must continue the request. This field will not be set or will be set
			# to 0 in the following cases:
			# (a) When the file_size is less than 512 KB.
			# (b) When the complete file has been covered by the diff computation.
			if 'next_offset' in crt_info:
				next_offset = crt_info['next_offset']
			else:
				next_offset = 0
				
			file_size = crt_info['file_size']

			# collect crt_info
			crt_info_list.append( crt_info )

			if next_offset > 0:
				# use subprocess for the remaining
				start_offset = next_offset
				end_offset = file_size
				(crt_info_list_2, error_msg) = ntnx_get_crt_info_multiprocess(max_procs, ss_file_path, reference_ss_file_path, start_offset, end_offset, file_size)
				if (error_msg != None):
					return 0, 0, None, error_msg
				next_offset = 0
				for c2 in crt_info_list_2:
					crt_info_list.append( c2 )
		else:
			while next_offset > start_offset:

				start_offset = next_offset

				(crt_info, error_msg) = ntnx_get_crt_info(ss_file_path, reference_ss_file_path, start_offset)
				if (error_msg != None):
					return 0, 0, None, error_msg

				# Next_offset : This field specifies the offset from where the client
				# must continue the request. This field will not be set or will be set
				# to 0 in the following cases:
				# (a) When the file_size is less than 512 KB.
				# (b) When the complete file has been covered by the diff computation.
				if 'next_offset' in crt_info:
					next_offset = crt_info['next_offset']
				else:
					next_offset = 0
					
				file_size = crt_info['file_size']

				if use_alignment:
					# collect crt_info
					crt_info_list.append( crt_info )

				else:
					(current_disk_backup_size, error_msg) = get_total_change_size(crt_info, is_skip_zeros)
					if (error_msg != None):
						return 0, 0, None, error_msg
					disks_backup_size += current_disk_backup_size

					(crt_info_xml, error_msg) = get_xml_for_crt_info(crt_info, is_skip_zeros, crt_info_xml)
					if (error_msg != None):
						return 0, 0, None, error_msg

		if use_alignment:
			aligned_result = align_changed_regions(crt_info_list, is_skip_zeros)
			(crt_info_xml, error_msg) = get_xml_for_aligned_result(aligned_result, is_skip_zeros)
			if (error_msg != None):
				return 0, 0, None, error_msg
			(current_disk_backup_size, error_msg) = get_total_change_size(crt_info, is_skip_zeros)
			if (error_msg != None):
				return 0, 0, None, error_msg
			disks_backup_size += current_disk_backup_size

		capacity = file_size

		disk_xml = SubElement(disks_xml, 'disk')

		disk_name = disk_label   #gen_disk_name(index)
		disk_key = str(naming_index)

		SubElementWithText(disk_xml, 'key', disk_key)
		SubElementWithText(disk_xml, 'name', disk_name)
		SubElementWithText(disk_xml, 'label', disk_label)
		SubElementWithText(disk_xml, 'uuid', uuid_from_ss_file_path)
		SubElementWithText(disk_xml, 'capacity', str(capacity))
		SubElementWithText(disk_xml, 'changeTrackingSkipped', '0')
		SubElementWithText(disk_xml, 'changeTrackingEnabled', '1')

		disk_xml.append(crt_info_xml)

		# naming index does not include cdrom
		naming_index += 1

	return nbr_disks_included, disks_backup_size, disks_xml, error_msg

def cleanup(target_ip_addr, client_ip_addr_list, vm_uuid, ss_uuid_remove, ss_uuid_keep):
	# cleanup iscsi
	cleanup_iscsi( target_ip_addr, vm_uuid )

	# cleanup vg
	cleanup_vg( client_ip_addr_list, vm_uuid )

	# cleanup ss
	cleanup_ss( client_ip_addr_list, vm_uuid, ss_uuid_remove, ss_uuid_keep )

def do_backup():
	
	info_message = ''
	warning_message = ''

	vm_uuid = options.uuid
	curr_time = time.time() #datetime.now()

	# do not backup a new vm which is still being restored
	if is_running_uni_process_restoring_vm(vm_uuid):
		error_msg = "Not allowed to backup VM with a restore in progress: %s" % vm_uuid
		output_xml_for_error(error_msg)
		sys.exit(2)

	# get appliance IP addresses
	# TODO: handle multiple addresses
	# client_ip_addr_list = [ "192.168.134.116" ]
	(client_ip_addr_list, error_msg) = get_dpu_ip_addresses()
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't get appliance IP address.")
		output_xml_for_error(error_msg)
		sys.exit(2)
	dbg_msg("dpu ip_addrs: %s" % (client_ip_addr_list,) )

	## get vm info
	(vm_info, vm_info_xml, error_msg) = get_xml_for_vm_info_for_metadata()
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't get VM metadata xml.")
		output_xml_for_error(error_msg)
		sys.exit(2)

	# check for skipped disk
	if g_disk_warning_msg:
		warning_message = append_error_msg(warning_message, g_disk_warning_msg)

	## get vm hardware hash
	(hw_hash_string, error_msg) = gen_hardware_hash_string(vm_info)
	if error_msg != None:
		error_msg = append_error_msg(error_msg, "Couldn't get VM hardware hash.")
		output_xml_for_error(error_msg)
		sys.exit(2)
	vm_hardware_hash_xml = Element('vm_hardware_hash')
	vm_hardware_hash_xml.text = hw_hash_string

	# if incr, then validate with hardwarehash passed on command line
	if options.snapshot != None and options.hardwarehash != None:
		prev_hw_hash_string = options.hardwarehash
		if hw_hash_string != prev_hw_hash_string:
			dbg_msg('H/W hash passed from previous Backup is not the same as the calculated one.')
			error_msg = 'VM configuration has changed.  A full backup is required.'
			promote_next_reason = 'A VM configuration change has been detected.'
			output_xml_for_error_with_promote_next(error_msg, promote_next_reason)
			sys.exit(2)
		dbg_msg("H/W hash matches prior VM configuration.")

	# get cluster target ip
	(target_ip_addr, error_msg) = ntnx_get_cluster_data_ip_addr()
	if (error_msg != None):
		error_msg = append_error_msg( error_msg, "Couldn't get Nutanix cluster data IP address during preparation for backup.")
		output_xml_for_error(error_msg)
		sys.exit(2)

	# assure previous run was properly cleaned-up

	# cleanup iscsi
	cleanup_iscsi( target_ip_addr, vm_uuid )

	# cleanup vg
	cleanup_vg( client_ip_addr_list, vm_uuid )

	# cleanup ss
	cleanup_ss( client_ip_addr_list, vm_uuid, None, options.snapshot )

	## get id 1
	idemp_name_ss = gen_id_name( client_ip_addr_list[0], vm_uuid, curr_time, 'ss' )

	if 0:	# assure prior ID is deleted
		(ss_uuid, error_msg) = ntnx_get_idempotence_id(idemp_name_ss)
		if (error_msg != None):
			# TODO: check error code
			#output_xml_for_error(error_msg)
			#sys.exit(2)
			dbg_msg("idempotence id not found: %s" % idemp_name_ss)
		else:
			(result, error_msg) = ntnx_delete_idempotence_id(idemp_name_ss)
			if (error_msg != None):
				output_xml_for_error(error_msg)
				sys.exit(2)

	(ss_uuid, error_msg) = ntnx_create_idempotence_id(idemp_name_ss)
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't create snapshot ID during preparation for backup.")
		output_xml_for_error(error_msg)
		sys.exit(2)

	## get id 2, TODO: get both id's with one request
	idemp_name_vg = gen_id_name( client_ip_addr_list[0], vm_uuid, curr_time, 'vg' )

	if 0:	# assure prior ID is deleted
		(vg_uuid, error_msg) = ntnx_get_idempotence_id(idemp_name_vg)
		if (error_msg != None):
			# TODO: check error code
			#output_xml_for_error(error_msg)
			#sys.exit(2)
			dbg_msg("idempotence id not found: %s" % idemp_name_vg)
		else:
			(result, error_msg) = ntnx_delete_idempotence_id(idemp_name_vg)
			if (error_msg != None):
				output_xml_for_error(error_msg)
				sys.exit(2)
		

	# delay between getting idempotence IDs to possibly reduce chance of proxy error
	time.sleep( 1 )  # seconds

	(vg_uuid, error_msg) = ntnx_create_idempotence_id(idemp_name_vg)
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't create volume group ID during preparation for backup.")
		output_xml_for_error(error_msg)
		sys.exit(2)

	vg_name = idemp_name_vg
	ss_info = None

	## create snapshot
	ss_name = idemp_name_ss

	if (0):
		# for dev testing, delete prior snapshot
		(ss_info, error_msg) = ntnx_ss_get(ss_uuid)
		if (error_msg != None):
			dbg_msg("Could not get existing snapshot: %s" % ss_uuid)
			#output_xml_for_error(error_msg)
			#sys.exit(2)
		else:
			# for dev testing, delete prior snapshot
			(result, error_msg) = ntnx_ss_delete(ss_uuid)
			if (error_msg != None):
				dbg_msg("Could not delete existing snapshot: %s" % ss_uuid)
				#output_xml_for_error(error_msg)
				# continue
				#  sys.exit(2)

	if 1:
		previous_ss_uuid = None
		previous_ss_info = None
		if options.snapshot != None:
			# doing incremental backup
			previous_ss_uuid = options.snapshot

			(previous_ss_info, error_msg) = ntnx_ss_get(previous_ss_uuid)
			if (error_msg != None):
				dbg_msg("Could not get previous snapshot: %s" % previous_ss_uuid)
				error_msg += "\nCould not get previous snapshot: %s" % previous_ss_uuid
				error_msg += '\nA full backup is required.'
				promote_next_reason = 'Could not get previous snapshot.  A full backup is required.'
				output_xml_for_error_with_promote_next(error_msg, promote_next_reason)
				sys.exit(2)
			# validate state of snapshot
			if 'status' not in previous_ss_info \
			or 'state' not in previous_ss_info['status'] \
			or previous_ss_info['status']['state'] not in ['kComplete', 'COMPLETE']:
				dbg_msg("Previous snapshot state is not complete: %s" % previous_ss_uuid)
				error_msg = "Previous snapshot state is not complete: %s" % previous_ss_uuid
				error_msg += '\nA full backup is required.'
				promote_next_reason = 'Invalid previous snapshot.  A full backup is required.'
				output_xml_for_error_with_promote_next(error_msg, promote_next_reason)
				sys.exit(2)
				
	if 1:
		# create new snapshot
		add_teardown_entry_ss_uuid( ss_uuid )
		(ss_info, error_msg) = ntnx_ss_create(vm_uuid, ss_name, ss_uuid)
		if (error_msg != None):
			cleanup(target_ip_addr, client_ip_addr_list, vm_uuid, ss_uuid, previous_ss_uuid)
			error_msg = append_error_msg(error_msg, "Couldn't create snapshot during preparation for backup.")
			output_xml_for_error(error_msg)
			sys.exit(2)

		(ss_type_msg, ss_warn_msg) = get_snapshot_type_message(ss_info)
		if ss_type_msg != None:
			if len(info_message) > 0:
				info_message += '\n'
			info_message += ss_type_msg
		if (ss_warn_msg != None):
			if len(warning_message) > 0:
				warning_message += '\n'
			warning_message += ss_warn_msg

	else:
		# for dev testing, get existing snapshot
		(ss_info, error_msg) = ntnx_ss_get(ss_uuid)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)

	## get change region tracking
	dbg_msg("crt.get")
	ss_info_2 = ss_info
	ss_uuid_1 = previous_ss_uuid
	ss_info_1 = previous_ss_info
	disk_list_1 = None

	is_incr = 0

	#if (is_incr):
	if (ss_uuid_1):
		(ss_uuid_1, disk_list_1, error_msg) = ntnx_ss_info_to_vg_disk_list(ss_info_1)
		if error_msg != None:
			error_msg = append_error_msg(error_msg, "Couldn't get disk information from snapshot during preparation for backup.")
			cleanup(target_ip_addr, client_ip_addr_list, vm_uuid, ss_uuid, previous_ss_uuid)
			return error_msg

	(ss_uuid_2, disk_list_2, error_msg) = ntnx_ss_info_to_vg_disk_list(ss_info_2)
	if error_msg != None:
		error_msg = append_error_msg(error_msg, "Couldn't get disk information from snapshot during preparation for backup.")
		cleanup(target_ip_addr, client_ip_addr_list, vm_uuid, ss_uuid, previous_ss_uuid)
		return error_msg

	disks_backup_size = 0
		
	# get xml for disks crt
	(nbr_disks_included, disks_backup_size, disks_xml, error_msg) = get_xml_for_disks( vm_info, disk_list_1, disk_list_2 )
	if error_msg != None:
		error_msg = append_error_msg(error_msg, "Couldn't get disk changed regions during preparation for backup.")
		cleanup(target_ip_addr, client_ip_addr_list, vm_uuid, ss_uuid, previous_ss_uuid)
		return error_msg

	backup_size_in_mb = disks_backup_size / 1024 / 1024
	dbg_msg("Current Backup size in MBs is - %s" % backup_size_in_mb)

	# if no disks with data to copy, then avoid creatng volume group and attaching with iscsi 
	iscsi_disk_devices_xml = None
	if nbr_disks_included > 0:

		# create volume group
		add_teardown_entry_vg_uuid( vg_uuid )
		(result, error_msg) = ntnx_create_volume_group(ss_info, vg_name, vg_uuid, client_ip_addr_list)
		if (error_msg != None):
			error_msg = append_error_msg( error_msg, "Failed to create volume group.")
			output_xml_for_error(error_msg)
			sys.exit(2)


		dbg_msg("iscsi.up")

		# columns: ver|ppid|lstart|tries|isci_detach|target_ip_addr|vm_uuid
		add_teardown_entry_iscsi( target_ip_addr, vm_uuid )

		(iscsi_disk_devices, error_msg) = iscsi_attach(target_ip_addr, vg_uuid)
		if (error_msg != None):
			error_msg = append_error_msg( error_msg, "Failed to create iscsi session.") 
			output_xml_for_error(error_msg)
			sys.exit(2)
		iscsi_disk_devices_xml = get_xml_from_iscsi_disk_devices(iscsi_disk_devices)


	root_xml = Element('aprotect')
	root_xml.append(vm_info_xml)
	root_xml.append(vm_hardware_hash_xml)

	if (info_message != None) and len(info_message) > 0:
		aprotect_info_xml = SubElement(root_xml, 'aprotect_info')
		SubElementWithText(aprotect_info_xml, 'info_message', info_message)

	if (warning_message != None) and len(warning_message) > 0:
		aprotect_warning_xml = SubElement(root_xml, 'aprotect_warning')
		SubElementWithText(aprotect_warning_xml, 'warning_message', warning_message)

	SubElementWithText(root_xml, 'backup_start_time', str(curr_time))

	SubElementWithText(root_xml, 'snapshot_name', ss_name)
	SubElementWithText(root_xml, 'snapshot_uuid', ss_uuid_2)

	SubElementWithText(root_xml, 'volume_group_name', vg_name)
	SubElementWithText(root_xml, 'volume_group_uuid', vg_uuid)

	SubElementWithText(root_xml, 'backupSizeMb', str(backup_size_in_mb))

	if disks_xml != None:
		root_xml.append(disks_xml)
	if iscsi_disk_devices_xml != None:
		root_xml.append(iscsi_disk_devices_xml)

	# dump xml
	indent(root_xml)
	ElementTree.dump(root_xml)

	return None

def delete_old_snapshots_for_vm( vm_uuid, client_ip_addr_list, keepsnapshot ):

	dbg_msg("Entering delete_old_snapshots_for_vm")

	if vm_uuid == None:
		dbg_msg("Can't delete old snapshots for vm. No vm_uuid specified")
		return 0, None

	(ss_list_result, error_msg) = ntnx_ss_list(vm_uuid)
	if (error_msg != None):
		#output_xml_for_error(error_msg)
		dbg_msg("Couldn't list snapshots: %s" % error_msg)
		return 0, None

	total_matches = ss_list_result['metadata']['total_matches']
	dbg_msg("found %d snapshots for vm_uuid %s" % (total_matches, vm_uuid))

	entities = ss_list_result['entities']
	for entity in entities:
		kind = entity['metadata']['kind']
		ss_uuid = entity['metadata']['uuid']
		name = entity['status']['name']
		if ss_uuid == keepsnapshot:
			dbg_msg("Keeping snapshot name=%s ss_uuid=%s" % (name,ss_uuid))
			continue

		# current example ss name: 
		# ub:192.168.134.116:84991b70-866d-4ab2-aec8-06892e145809:0:ss
		# name_ss = gen_id_name( client_ip_addr_list[0], vm_uuid, curr_time, 'ss' )

		name_parts = name.split(':')
		if len(name_parts) < 3:
			dbg_msg("Skipping name=%s ss_uuid=%s" % (name,ss_uuid))
			continue
		if name_parts[0] != 'ub' \
		or name_parts[1] != client_ip_addr_list[0] \
		or name_parts[2] != vm_uuid:
			dbg_msg("Skipping snapshot name=%s ss_uuid=%s" % (name,ss_uuid))
			continue

		dbg_msg("Deleting snapshot name=%s ss_uuid=%s" % (name,ss_uuid))
		(result, error_msg) = ntnx_ss_delete(ss_uuid)
		if (error_msg != None):
			# TODO: just a warning
			dbg_msg("Couldn't delete snapshot: %s " % (error_msg))
			#output_xml_for_error(error_msg)
			#sys.exit(2)

	dbg_msg("Leaving delete_old_snapshots_for_vm")

	return 0, None

def cleanup_snapshot_by_ss_uuid(deletion_ss_uuid):
	error_msg = None
	if (deletion_ss_uuid == None):
		error_msg = "deletion_ss_uuid is None"
		return None, error_msg
	dbg_msg('cleanup_snapshot_by_ss_uuid: ' + deletion_ss_uuid)

	(previous_ss_info, error_msg) = ntnx_ss_get(deletion_ss_uuid)
	if (error_msg != None):
		dbg_msg("Could not get deletion snapshot: %s" % deletion_ss_uuid)
		if 'Not found' in error_msg:
			dbg_msg("deletion snapshot was not found: %s" % deletion_ss_uuid)
			# if not found, then nothing to delete. return success
			return None, None
	else:
		# delete snapshot
		(result, error_msg) = ntnx_ss_delete(deletion_ss_uuid)
		if (error_msg != None):
			dbg_msg(error_msg)
			if 'Not found' in error_msg:
				# if not found, then nothing to delete
				return None, None
			return None, error_msg

	return None, error_msg

def cleanup_volume_group(vg_uuid):
	if (vg_uuid == None):
		return
	dbg_msg('cleanup_volume_group: ' + vg_uuid)
	if vg_uuid != None:
		(vg_info, error_msg) = ntnx_get_volume_group(vg_uuid)
		if (error_msg != None):
			dbg_msg("Could not get volume group: %s" % vg_uuid)
			# keep going
		else:
			vg_name = vg_info['name']

			# close volume group
			(result, error_msg) = ntnx_close_volume_group(vg_info, vg_name, vg_uuid)
			if (error_msg != None):
				dbg_msg("Could not close volume group: %s" % vg_uuid)
				# keep going

			# delete volume group
			(result, error_msg) = ntnx_delete_volume_group(vg_name, vg_uuid)
			if (error_msg != None):
				dbg_msg("Could not delete volume group: %s" % vg_uuid)
				# keep going
	return

def delete_old_volume_groups(client_ip_addr_list, vm_uuid ):
	dbg_msg("Entering cleanup_old_volume_groups")

	# delete any volume group with matching name format (ip, vm_uuid)
	(result, error_msg) = ntnx_get_volume_groups()
	if (error_msg != None):
		dbg_msg("Could not get volume groups:")
		return

	count = result['metadata']['count']
	entities = result['entities']
	for index in range(0,count):
		entity = entities[index]

		vg_name = entity['name']
		vg_uuid = entity['uuid']

		# filter out those of interest
		# current example vg name: 
		# ub:192.168.134.116:84991b70-866d-4ab2-aec8-06892e145809:0:vg
		# name_ss = gen_id_name( client_ip_addr_list[0], vm_uuid, curr_time, 'vg' )

		name = vg_name
		name_parts = name.split(':')
		if len(name_parts) < 3:
			dbg_msg("Skipping volume group name=%s vg_uuid=%s" % (name, vg_uuid))
			continue
		if name_parts[0] != 'ub' \
		or name_parts[1] != client_ip_addr_list[0] \
		or name_parts[2] != vm_uuid:
			dbg_msg("Skipping volume group name=%s vg_uuid=%s" % (name, vg_uuid))
			continue

		# close volume group
		(result, error_msg) = ntnx_close_volume_group(entity, vg_name, vg_uuid)
		if (error_msg != None):
			dbg_msg("Couldn't close volume group name=%s vg_uuid=%s" % (name, vg_uuid))
			# keep going

		# delete volume group
		dbg_msg("Deleting volume group name=%s vg_uuid=%s" % (name,vg_uuid))
		(result, error_msg) = ntnx_delete_volume_group(vg_name, vg_uuid)
		if (error_msg != None):
			dbg_msg("Could not delete volume group name=%s vg_uuid=%s" % (name, vg_uuid))
			# keep going

	dbg_msg("Leaving cleanup_old_volume_groups")
	return

def cleanup_iscsi(target_ip_addr, vm_uuid):
	# iscsi down
	dbg_msg("iscsi.down")
	(result, error_msg) = iscsi_detach(target_ip_addr, vm_uuid)
	if (error_msg != None):
		dbg_msg(error_msg)
		# keep going

def cleanup_vg_by_vg_uuid(vg_uuid):

	(vg_info, error_msg) = ntnx_get_volume_group(vg_uuid)
	if (error_msg != None):
		dbg_msg("Could not get volume group %s: %s" % (vg_uuid, error_msg))
		# check for known error
		ix = error_msg.find('Unknown vgs:')
		if ix > -1:
			# Couldn't find the volume group. Nothing more to do.
			return None, None
		return None, error_msg

	result_error_msg = None
	vg_name = vg_info['name']

	(result, error_msg) = ntnx_close_volume_group(vg_info, vg_name, vg_uuid)
	if (error_msg != None):
		dbg_msg("Could not close volume group %s: %s", (vg_uuid, error_msg))
		if len(result_error_msg) > 0:
			result_error_msg += '\n'
		result_error_msg += error_msg
		# continue

	(result, error_msg) = ntnx_delete_volume_group(vg_name, vg_uuid)
	if (error_msg != None):
		dbg_msg("Could not delete volume group %s: %s", (vg_uuid, error_msg))
		if len(result_error_msg) > 0:
			result_error_msg += '\n'
		result_error_msg += error_msg
		# continue

	(result, error_msg) = ntnx_delete_idempotence_id(vg_name)
	if (error_msg != None):
		dbg_msg("Could not delete idempotence_id %s: %s", (vg_name, error_msg))
		if len(result_error_msg) > 0:
			result_error_msg += '\n'
		result_error_msg += error_msg
		# continue

	return None, result_error_msg

def TEST_cleanup_idemp_id(idemp_name):
	(id, error_msg) = ntnx_get_idempotence_id(idemp_name)
	if (error_msg != None):
		dbg_msg("Couldn't find idempotence id: %s" % idemp_name)
		return None, error_msg
	else:
		# delete idempotence id
		(result, error_msg) = ntnx_delete_idempotence_id(idemp_name)
		if (error_msg != None):
			dbg_msg("Could not delete idempotence id: %s" % idemp_name_vg)
	return None, None

def cleanup_vg(client_ip_addr_list, vm_uuid):
	# get vg id
	idemp_name_vg = gen_id_name( client_ip_addr_list[0], vm_uuid, 0, 'vg' )
	(vg_uuid, error_msg) = ntnx_get_idempotence_id(idemp_name_vg)
	if (error_msg != None):
		dbg_msg("Couldn't find idempotence id: %s" % idemp_name_vg)
		# keep going
	else:
		# delete volume group
		# TODO: do we need this cleanup here?
		# the delete_old_volume_groups below will get rid of this one.
		# cleanup_volume_group(vg_uuid)

		# delete idempotence id
		(result, error_msg) = ntnx_delete_idempotence_id(idemp_name_vg)
		if (error_msg != None):
			dbg_msg("Could not delete idempotence id: %s" % idemp_name_vg)

	# delete any volume group with matching name format (ip, vm_uuid)
	delete_old_volume_groups(client_ip_addr_list, vm_uuid )


def TEST_cleanup_ss_id(client_ip_addr_list, vm_uuid):
	# get ss id
	idemp_name_ss = gen_id_name( client_ip_addr_list[0], vm_uuid, 0, 'ss' )
	(ss_uuid, error_msg) = ntnx_get_idempotence_id(idemp_name_ss)
	if (error_msg != None):
		dbg_msg("Couldn't find idempotence id: %s" % idemp_name_ss)
		# keep going

	# delete idempotent ID
	(result, error_msg) = ntnx_delete_idempotence_id(idemp_name_ss)
	if (error_msg != None):
		dbg_msg("Could not delete idempotence id: %s" % idemp_name_ss)

	return

def cleanup_ss(client_ip_addr_list, vm_uuid, delsnapshot, keepsnapshot):
	# get ss id
	idemp_name_ss = gen_id_name( client_ip_addr_list[0], vm_uuid, 0, 'ss' )
	(ss_uuid, error_msg) = ntnx_get_idempotence_id(idemp_name_ss)
	if (error_msg != None):
		dbg_msg("Couldn't find idempotence id: %s" % idemp_name_ss)
		# keep going
	
	if delsnapshot == None:
		# cleanup any leftover snapshots for the vm,
		# but only if there is a specific snapshot to keep
		if keepsnapshot != None:
			delete_old_snapshots_for_vm( vm_uuid, client_ip_addr_list, keepsnapshot )
	else:
		# cleanup specific snapshot
		cleanup_snapshot_by_ss_uuid(delsnapshot)

	## cleanup any leftover snapshots
	#if options.keepsnapshot != None:
	#	delete_old_snapshots_for_vm( vm_uuid, client_ip_addr_list, options.keepsnapshot )

	# delete idempotent ID
	(result, error_msg) = ntnx_delete_idempotence_id(idemp_name_ss)
	if (error_msg != None):
		dbg_msg("Could not delete idempotence id: %s" % idemp_name_ss)

	return

def do_postbackup():
	dbg_msg("Entering do_postbackup")

	warning_msg = ""
	info_msg = ""

	vm_uuid = options.uuid
	vg_uuid = options.volumegroup
	curr_time = datetime.datetime.now()
	start_time = options.start_time

	# get appliance IP addresses
	(client_ip_addr_list, error_msg) = get_dpu_ip_addresses()
	if (error_msg != None):
		error_msg = append_error_msg(error_msg, "Couldn't get appliance IP address.")
		output_xml_for_error(error_msg)
		sys.exit(2)
	dbg_msg("dpu ip_addrs: %s" % (client_ip_addr_list,) )

	(target_ip_addr, error_msg) = ntnx_get_cluster_data_ip_addr()
	if (error_msg != None):
		error_msg = append_error_msg( error_msg, "Couldn't get Nutanix cluster data IP address to finish backup.")
		output_xml_for_error(error_msg)
		sys.exit(2)

	# cleanup iscsi
	cleanup_iscsi( target_ip_addr, vm_uuid )
	remove_teardown_entry_iscsi( target_ip_addr, vm_uuid )

	# cleanup vg
	cleanup_vg( client_ip_addr_list, vm_uuid )
	if vg_uuid != None:
		remove_teardown_entry_vg_uuid( vg_uuid )

	# cleanup ss
	cleanup_ss( client_ip_addr_list, vm_uuid, options.delsnapshot, options.keepsnapshot )
	# remove ss teardown entry
	ss_uuid = options.snapshot
	if ss_uuid != None:
		remove_teardown_entry_ss_uuid( ss_uuid )

	# build xml result
	root_aprotect = Element('aprotect')
	child_aprotect = SubElement(root_aprotect, 'postbackup_result')
	child_aprotect_error_msg = SubElement(child_aprotect, 'status')
	child_aprotect_error_msg.text = 'success'

	# dump xml
	indent(root_aprotect)
	ElementTree.dump(root_aprotect)

	dbg_msg("Leaving do_postbackup")

def prep_vm_nics_info_from_et(vm_nics_et):
	vm_nics = []
	for vm_nic_et in vm_nics_et.getchildren():
		vm_nic = {}
		for r in vm_nic_et.getchildren():
			if r.tag in [ 'network_uuid', 'model' ]:
				text = ""
				if r.text != None:
					text = r.text
				vm_nic[ r.tag ] = text
		vm_nics.append(vm_nic)
	return vm_nics

def prep_boot_info_from_et(tree):
	boot_info = {}

	boot_device_type_et = tree.find('boot_device_type')
	if boot_device_type_et != None:
		boot_info['boot_device_type'] = boot_device_type_et.text

	disk_address_dict = {}
	disk_address_et = tree.find('disk_address')
	if disk_address_et != None:
		disk_address_dict['device_index'] = int( disk_address_et.find('device_index').text )
		disk_address_dict['device_bus'] = disk_address_et.find('device_bus').text
		boot_info['disk_address'] = disk_address_dict

	return boot_info


#	et2data_dict
#	{	'disk_address':		{ type:'dict', {'device_index':	'i', 
#			  				'disk_label':	's',
#			  				'device_bus':	's' } ),
#		'is_cdrom':		{ type:'b' }
#		'is_scsi_passthrough':	{ type:'b' }
#		'is_empty':		'b',
#		'flash_mode_enabled':	'b',
#		'shared':		'b',
#		'is_thin_provisioned':	'b',
#		'size':			'i',
#	}
#	add_item( 'is_cdrom', 'b' )
#	add_item( 'is_scsi_passthrough', 'b')
#	add_item( 'is_empty', 'b')
#
#	for child in tree:
#		xlate = vm_disk_et2data_xlate
#		if child.tag in xlate
#			r = xlate[child.tag]
#			type = r['type']
#			if type == 'b':
#				if child.text == "True":
#					bval = True
#				else:
#					bval = False
#				data[child.tag] = bval
#			elif type == 'i':
#				data[child.tag] = int(child.text)
#			elif type == 's':
#				data[child.tag] = child.text
#			elif type == 'dict':
#				data[child.tag] = 

def prep_vm_disks_from_et(tree, diskindexmap=None):
	vm_disks = []
	index = 0
	for disk_et in tree:

		is_excluded = False
		disk = {}
		for child in disk_et:

			if child.tag == 'disk_address':
				disk_address = {}
				disk_address_et = child.getchildren()
				for da in disk_address_et:  
					if da.tag in [ 'disk_label', 'device_bus', 'vmdisk_uuid' ]:
						disk_address[ da.tag ] = da.text
					if da.tag in [ 'device_index' ]:
						disk_address[ da.tag ] = int(da.text)
				disk['disk_address'] = disk_address

				#"vm_disk_clone": {
				#	"disk_address": disk_address
				#		"device_bus": "SCSI",
				#		"device_index": 0,
				#		#"ndfs_filepath": "string",
				#		#"vmdisk_uuid": "string",
				#		#"volume_group_uuid": "string"
				#	}
				#disk['vm_disk_clone'] = vm_disk_clone

			elif child.tag in ['is_cdrom', 'is_scsi_passthrough', 'is_empty',
					'flash_mode_enabled', 'shared', 'is_thin_provisioned']:
				if child.text == "True":
					bval = True
				else:
					bval = False
				disk[child.tag] = bval
				
			elif child.tag in ['size']:
				disk[child.tag] = int(child.text)
				the_size = int(child.text)

			elif child.tag in ['is_excluded']:
				is_excluded = get_et_boolean(child)
				disk[child.tag] = True

				#vm_disk_create = {
				#		'size': the_size,
				#		'storage_container_uuid': options.storagecontainer
				#		 }
				#disk['vm_disk_create'] = vm_disk_create

		vm_disks.append(disk)

		if diskindexmap != None:
			disk_label = disk['disk_address']['disk_label']
			dbg_msg( "disk_label is %s" % disk_label )
			diskindexmap.append( { 
				'key': str(index),
				'name': disk_label,
				'is_excluded': is_excluded
				} )
			index += 1

	return vm_disks

def prep_cdrom_disk_from_prior_vm_disks(prior_vm_disks):
	for prior_disk in prior_vm_disks:
		if prior_disk['is_cdrom']:
			vm_disk_address = prior_disk['disk_address']
			disk_address = { 
				"device_bus": vm_disk_address['device_bus'],
				"device_index": vm_disk_address['device_index']
				}
			disk = {}
			disk['disk_address'] = disk_address
			disk['is_cdrom'] = True

			vm_disk_info = [ disk ]
			return vm_disk_info

	return None

def get_vg_info_disk_for_index(vg_info, index):
	vg_disk_list = vg_info['disk_list']
	for disk in vg_disk_list:
		disk_index = disk['index']
		if disk_index == index:
			return disk
	return None

def prep_vm_disks_for_attach(prior_vm_disks, diskindexmap, vg_info):
	error_msg = None
	result = []
	storagecontainer = options.storagecontainer
	
	vg_disk_index = 0

	dbg_msg("AAA")
	for ix in range(0,len(prior_vm_disks)):
		dbg_msg("CCC check %d" % ix)
		vm_disk = prior_vm_disks[ix]
		dbg_msg("CCC vm_disk %s" % vm_disk)

		if 'is_excluded' in vm_disk:
			is_excluded = vm_disk['is_excluded']
		else:
			is_excluded = False

		if options.metadata != None:
			dbg_msg("options.metadata=%s" % options.metadata)

		#vg_disk = get_vg_info_disk_for_index(vg_info, ix)
		#if vg_disk == None:
		#	dbg_msg("Could not find disk in volume group for index %d" % ix)
		#	##return None, error_msg
		#	continue

		vm_disk_address = vm_disk['disk_address']

		disk_address = { 
				"device_bus": vm_disk_address['device_bus'],
				"device_index": vm_disk_address['device_index']
				#"device_bus": "IDE",
				#"device_index": 1,
				#"vmdisk_uuid": vg_disk['vmdisk_uuid']
				#"volume_group_uuid": "string"
				}

		disk = {}

		if vm_disk['is_cdrom']:
			disk['disk_address'] = disk_address
			disk['is_cdrom'] = True
			disk['is_empty'] = True
			result.append( disk )

		elif is_excluded:
			if (options.metadata == None or options.metadata.upper() != 'NO'):
				# restore metadata of excluded disks,
				# i.e. create empty disk of appropriate size
				disk['disk_address'] = disk_address
				vm_disk_create = {
					"size": vm_disk['size'],
					"storage_container_uuid": storagecontainer
					}
				disk['vm_disk_create'] = vm_disk_create
				###vg_disk_index += 1
				result.append( disk )

		else:
			# only add disks with vmdisk_uuid
			if 'vmdisk_uuid' in vm_disk_address:
				# restore disk
				disk['disk_address'] = disk_address
				vg_disk = vg_info['disk_list'][vg_disk_index]
				#dbg_msg("Found disk in volume group for index %d" % ix )
				vm_disk_clone = {
					"disk_address": {
						#"device_bus": "IDE",
						#"device_index": 1,
						"vmdisk_uuid": vg_disk['vmdisk_uuid']
						#"volume_group_uuid": "string"
						}
					}
				disk['vm_disk_clone'] = vm_disk_clone
				vg_disk_index += 1
				result.append( disk )
		
	dbg_msg("prep_vm_disks_for_attach: prior_vm_disks %s" % prior_vm_disks)
	return result, error_msg

def prep_create_vm_info_from_et(tree):

	new_vm_info = {}

	for child in tree:  
		if child.tag in [ 'name', 'timezone' ]:
			new_vm_info[ child.tag ] = child.text

		if child.tag in [ 'num_cores_per_vcpu', 'num_vcpus',
				 'memory_mb', 'ha_priority' ]:
			new_vm_info[ child.tag ] = int(child.text)

		if child.tag in [ 'gpus_assigned', 'allow_live_migration' ]:
			if child.text == "True":
				bval = True
			else:
				bval = False
			new_vm_info[ child.tag ] = bval
			
		#todo: power_state

		if 0 and child.tag in [ 'disks' ]:
			new_vm_info['vm_disks'] = prep_vm_disks_from_et(child)

		#if child.tag in [ 'boot' ]:
		#	new_vm_info[ child.tag ] = prep_boot_info_from_et(child)

		#if child.tag in [ 'vm_nics' ]:
		#	new_vm_info[ child.tag ] = prep_vm_nics_info_from_et(child)
			
	json_print = json.dumps(new_vm_info, indent = 4, separators = (',', ': '))
	dbg_msg("do_restore_vm_prep: what we got:\n" + json_print )
	return new_vm_info

def prep_disk_index_to_key(tree):
	disk_index_to_key = []
	index = 0

	disks_et = tree.find('disks')
	if disks_et:
		for disk_et in disks_et.getchildren():
			key_et = disk_et.find('key')
			key = int(key_et.text)
			disk_index_to_key.append(key)
	return disk_index_to_key

def prep_create_vg_disk_list_from_et(tree, diskindexmap=None):
	storage_container_uuid = options.storagecontainer
	error_msg = None
	disk_list = []
	index = 0

	disks_et = tree.find('disks')
	if disks_et:
		#vm_disks = prep_vm_disks_from_et(child)
		#dbg_msg("vm_disks %s" % vm_disks)
		#return prep_create_vg_disk_list_from_vm_disks(vm_disks)

		for disk_et in disks_et.getchildren():

			# skip cdrom
			if disk_et.tag == 'cdrom':
				continue

			# process disk
			disk = {}

			name_et = disk_et.find('name')
			name = name_et.text

			key_et = disk_et.find('key')
			key = int(key_et.text)	# use the key as he index value

			is_excluded_et = disk_et.find('is_excluded')
			if is_excluded_et:
				is_excluded = get_et_boolean(is_excluded_et)
			else:
				is_excluded = False
				
			disk_address_et = disk_et.find('disk_address')
			if disk_address_et != None and not is_excluded:
				vmdisk_uuid_et = disk_address_et.find('vmdisk_uuid')
				if vmdisk_uuid_et != None:
					# only add disk if it has vm_disk_uuid and not excluded

					vmdisk_uuid = vmdisk_uuid_et.text

					size_et = disk_et.find('size')
					if size_et != None:
						size = int(size_et.text)
					else:
						size = 0

					# be sure size is a multiple of MB because it is truncated to MB by the API.
					if ((size % ONE_MB) != 0):
						dbg_msg("specified disk size not multiple of MB: %d" % size)
						size = ((size + ONE_MB - 1) / ONE_MB) * ONE_MB
						dbg_msg("disk size rounded to next MB: %d" % size)

					disk_list.append( {
						"create_config": {
							"size": size,
							"storage_container_uuid": storage_container_uuid },
						"index": index
						} )
					index += 1

					if diskindexmap != None:
						diskindexmap.append( {
							'key': str(key),
							'name': name
							} )

	return disk_list, None

def OLD_prep_create_vg_disk_list_from_et(tree):
	vm_disks = []
	for disk_et in tree:

		disk = {}
		for child in disk_et:

			if child.tag == 'disk_address':
				disk_address = {}
				disk_address_et = child.getchildren()
				for da in disk_address_et:  
					if da.tag in [ 'disk_label', 'device_bus', 'vmdisk_uuid' ]:
						disk_address[ da.tag ] = da.text
					if da.tag in [ 'device_index' ]:
						disk_address[ da.tag ] = int(da.text)
				disk['disk_address'] = disk_address

			elif child.tag in ['is_cdrom', 'is_scsi_passthrough', 'is_empty',
					'flash_mode_enabled', 'shared', 'is_thin_provisioned']:
				if child.text == "True":
					bval = True
				else:
					bval = False
				disk[child.tag] = bval
				
			elif child.tag in ['size']:
				disk[child.tag] = int(child.text)
				the_size = int(child.text)

		vm_disks.append(disk)

	return vm_disks

def OLD_prep_create_vg_disk_list_from_vm_disks(vm_disks):
	# build vg disk list from vm_info
	error_msg = None
	storage_container_uuid = options.storagecontainer
	disk_list = []
	try:
		#vm_disk_info = vm_info['vm_disk_info']
		count = 0
		#for vm_disk in vm_disk_info:
		for vm_disk in vm_disks:
			size = vm_disk['size']
			disk_address = vm_disk['disk_address']
			is_cdrom = vm_disk['is_cdrom']
			# only add disk if it has vm_disk_uuid
			#if not is_cdrom:
			if 'vmdisk_uuid' in disk_address:
				##vmdisk_uuid = disk_address['vmdisk_uuid']
				##disk_address_item = { "vmdisk_uuid" : vmdisk_uuid }
				##disk_list.append( {
				##	"clone_config": { "disk_address": disk_address_item },
				##	"index": count
				##	} )
				disk_list.append( {
					"create_config": {
						"size": size,
						"storage_container_uuid": storage_container_uuid },
					"index": count
					} )
			count = count + 1
	except:
		dbg_msg("Failed to get disk list")
		return None, "Failed to get disk list"

	return disk_list, error_msg

def get_vm_disks_xml_from_vg_disk_list(disk_list, diskindexmap):
	disks_xml = Element('disks')
	for ix in range(0, len(disk_list)):
		##vmdisk_uuid = disk_list[ix]['clone_config']['disk_address']['vmdisk_uuid']
		disk_xml = SubElement(disks_xml, 'disk')
		#SubElementWithText(disk_xml, 'key', str(ix) )
		SubElementWithText(disk_xml, 'key', diskindexmap[ix]['key'] )
		SubElementWithText(disk_xml, 'name', diskindexmap[ix]['name'] )
		##SubElementWithText(disk_xml, 'uuid', vmdisk_uuid )
		##disk_address_xml = SubElement(disk_xml, 'disk_address')
		##SubElementWithText(disk_address_xml, 'vmdisk_uuid', vmdisk_uuid)
	return disks_xml

def add_nic_to_vm(vm_uuid, nic_info):
	error_msg = None

	# add nic to vm
	url = gen_url_ntnx_api_v2() + "/" + CONST_VMS + "/" + vm_uuid + "/" + CONST_NICS
	body = {
		'spec_list': [ nic_info ] 
	}

	# request with wait for task completion
	(result, error_msg) = rest_api_with_task_completion_with_retry(url, body)
	if (error_msg != None):
		return (None, error_msg)

	return (result, error_msg)

def find_network_uuid_from_all_networks_info(all_networks_info, vlan_id):
	total_entities = all_networks_info['metadata']['total_entities']
	entities = all_networks_info['entities']
	for ix in range(0,total_entities):
		entity = entities[ix]
		if entity['vlan_id'] == vlan_id:
			return entity['uuid']
	return None

def add_nics_to_vm_from_et(vm_uuid, metadata_et):
	dbg_msg("Entering add_nics_to_vm_from_et")
	warning_message = None

	# get all networks info
	all_networks_info, error_msg = ntnx_get_network_info("")  # empty string for network_uuid to get all networks
	if error_msg != None:
		dbg_msg("Couldn't get Nutanix cluster network info list.")
		return None, error_msg

	vm_nics_et = metadata_et.find('vm_nics')
	if vm_nics_et:
		vm_nics = []
		for vm_nic_et in vm_nics_et.getchildren():
			vm_nic = {}
			for r in vm_nic_et.getchildren():
				# TODO: handle requested_ip_address
				# if r.tag in [ 'model', 'request_ip', 'requested_ip_address' ]:
				if r.tag in [ 'model' ]:
					if r.text != None:
						vm_nic[ r.tag ] = r.text

			# use vlan_id to find network_uuid
			vlan_id = 0
			vlan_id_et = vm_nic_et.find('vlan_id')
			if vlan_id_et != None:
				dbg_msg("vlan_id.text: %s" % vlan_id_et.text)
				vlan_id = int( vlan_id_et.text )
			else:
				dbg_msg("no vlan_id")
				continue

			# lookup network_uuid in networks using vlan_id
			dbg_msg("look for vlan_id %d" % vlan_id)
			network_uuid = uuid = find_network_uuid_from_all_networks_info(all_networks_info, vlan_id)
			if network_uuid != None:
				vm_nic['network_uuid'] = network_uuid

				# add nic to the vm
				result, error_msg = add_nic_to_vm(vm_uuid, vm_nic)
				if error_msg != None:
					# treat this as a warning
					msg = "Couldn't add NIC to VM for vlan_id %d." % (vlan_id)
					msg = append_error_msg(error_msg, msg)
					dbg_msg(msg)
					warning_message = append_error_msg(warning_message, msg)
			else:
				msg = "Couldn't find matching cluster network with vlan_id %d." % (vlan_id)
				dbg_msg(msg)
				warning_message = append_error_msg(warning_message, msg)

	dbg_msg("Leaving add_nics_to_vm_from_et")
	return warning_message, None

def do_restore():
	dbg_msg("Entering do_restore")
	dbg_msg( "options.newvmname         : %s" % options.newvmname )
	dbg_msg( "options.storagecontainer  : %s" % options.storagecontainer )
	dbg_msg( "options.xml_filename      : %s" % options.xml_filename )

	cleanup_on_error = True
	info_message = ''
	warning_message = ''

	vg_name = None
	vg_uuid = None
	
	if 0:
		f = sys.stdin
	else:
		f = open(options.xml_filename, "r")

	# The metadata xml is on stdin
	# This try except is from
	# https://stackoverflow.com/questions/34143685/correct-encoding-for-xml-file-read-from-stdin
	try:
		input = f.buffer
	except AttributeError:
		input = f
	xmlString = input.read()
	##doc = xml.dom.minidom.parseString(xmlString)

	metadata_et = ElementTree.fromstring(xmlString)
	#print "%s : %s : %s\n" % (tree.tag, tree.attrib, tree.text)
	#for child in tree:  
	#	dbg_msg("%s : %s\n" % (child.tag, child.text) )
	#	print "%s : %s" % (child.tag, child.text)

	disk_index_to_key = prep_disk_index_to_key(metadata_et)

	vm_info = prep_create_vm_info_from_et(metadata_et)
	vm_info['name'] = options.newvmname

	# create vm
	vm_uuid, error_msg = ntnx_vm_create(vm_info)
	if (error_msg != None):
		error_msg = append_error_msg( error_msg, "Couldn't create VM.")
		output_xml_for_error(error_msg)
		sys.exit(2)

	# add nics to the vm
	nics_warning_message, error_msg = add_nics_to_vm_from_et(vm_uuid, metadata_et)
	if (error_msg != None):
		error_msg = append_error_msg( error_msg, "Couldn't add NIC to VM.")
		output_xml_for_error(error_msg)
		sys.exit(2)
	if (nics_warning_message != None):
		warning_message = append_error_msg(warning_message, nics_warning_message)

	# set vm_uuid in options.
	# other functions will look at this
	options.uuid = vm_uuid

	vm_info, error_msg = ntnx_vm_get_info(vm_uuid)
	if (error_msg != None):
		error_msg = append_error_msg( error_msg, "Couldn't get information for new VM.")
		output_xml_for_error(error_msg)
		sys.exit(2)

	# add vm_restoring entry to info file
	add_teardown_entry_vm_restoring( vm_uuid )

	diskindexmap = []
	vg_disk_list, error_msg = prep_create_vg_disk_list_from_et(metadata_et, diskindexmap)
	dbg_msg("vg_disk_list %s" % (vg_disk_list) )

	# if no disks with data to copy, then avoid creatng volume group and attaching with iscsi 
	iscsi_disk_devices_xml = None
	nbr_disks_included = len(vg_disk_list)
	if nbr_disks_included > 0:

		# get appliance IP addresses
		(client_ip_addr_list, error_msg) = get_dpu_ip_addresses()
		if (error_msg != None):
			error_msg = append_error_msg( error_msg, "Couldn't get appliance IP address.")
			output_xml_for_error(error_msg)
			sys.exit(2)
		dbg_msg("dpu ip_addrs: %s" % (client_ip_addr_list,) )

		# get idempotence id
		idemp_name_vg = gen_id_name( client_ip_addr_list[0], vm_uuid, 0, 'vg' )
		(vg_uuid, error_msg) = ntnx_create_idempotence_id(idemp_name_vg)
		if (error_msg != None):
			output_xml_for_error(error_msg)
			sys.exit(2)
		vg_name = idemp_name_vg

		# get cluster target ip
		(target_ip_addr, error_msg) = ntnx_get_cluster_data_ip_addr()
		if (error_msg != None):
			error_msg = append_error_msg( error_msg, "Couldn't get Nutanix cluster data IP address during preparation for restore.")
			output_xml_for_error(error_msg)
			sys.exit(2)

		# assure previous run was properly cleaned-up
		#cleanup_iscsi( target_ip_addr, vm_uuid )
		#cleanup_vg( client_ip_addr_list, vm_uuid )

		# create volume group
		add_teardown_entry_vg_uuid( vg_uuid )
		description = "volume group for restore of vm_uuid %s" % vm_uuid
		result, error_msg = ntnx_vg_create(vm_uuid, description, vg_disk_list, vg_name, vg_uuid, client_ip_addr_list)
		if (error_msg != None):
			error_msg = append_error_msg( error_msg, "Failed to create volume group.")
			output_xml_for_error(error_msg)
			sys.exit(2)

		dbg_msg("iscsi.up")
		add_teardown_entry_iscsi( target_ip_addr, vm_uuid )
		(iscsi_disk_devices, error_msg) = iscsi_attach(target_ip_addr, vg_uuid)
		if (error_msg != None):
			error_msg = append_error_msg( error_msg, "Failed to create iscsi session.") 
			if cleanup_on_error:
				cleanup_iscsi( target_ip_addr, vm_uuid )
				cleanup_vg( client_ip_addr_list, vm_uuid )
			output_xml_for_error(error_msg)
			sys.exit(2)
		iscsi_disk_devices_xml = get_xml_from_iscsi_disk_devices(iscsi_disk_devices, diskindexmap)

	# build xml result
	root_xml = Element('aprotect')

	if (info_message != None) and len(info_message) > 0:
		aprotect_info_xml = SubElement(root_xml, 'aprotect_info')
		SubElementWithText(aprotect_info_xml, 'info_message', info_message)

	if (warning_message != None) and len(warning_message) > 0:
		aprotect_warning_xml = SubElement(root_xml, 'aprotect_warning')
		SubElementWithText(aprotect_warning_xml, 'warning_message', warning_message)

	arestore_xml = SubElement(root_xml, 'aprotect_restore')

	vm_xml = SubElement(arestore_xml, 'virtual_machine')
	SubElementWithText(vm_xml, 'name', options.newvmname)
	SubElementWithText(vm_xml, 'uuid', vm_uuid)

	disks_xml = get_vm_disks_xml_from_vg_disk_list(vg_disk_list, diskindexmap)
	vm_xml.append(disks_xml)

	if vg_name == None:
		vg_name = "none"
	if vg_uuid == None:
		vg_uuid = "none"
	SubElementWithText(arestore_xml, 'volume_group_name', vg_name)
	SubElementWithText(arestore_xml, 'volume_group_uuid', vg_uuid)

	if iscsi_disk_devices_xml != None:
		root_xml.append(iscsi_disk_devices_xml)

	# dump xml
	indent(root_xml)
	ElementTree.dump(root_xml)

	dbg_msg("Leaving do_restore")

def do_postrestore():
	dbg_msg("Entering do_postrestore")
	dbg_msg( "options.uuid            : %s" % options.uuid )

	vm_uuid = options.uuid
	vg_uuid = None

	# get cluster target ip
	(target_ip_addr, error_msg) = ntnx_get_cluster_data_ip_addr()
	if (error_msg != None):
		error_msg = append_error_msg( error_msg, "Couldn't get Nutanix cluster data IP address to finish restore.")
		output_xml_for_error(error_msg)
		sys.exit(2)
	dbg_msg("cluster target ip_addr: %s" % target_ip_addr )

	# get appliance IP addresses
	(client_ip_addr_list, error_msg) = get_dpu_ip_addresses()
	if (error_msg != None):
		error_msg = append_error_msg( error_msg, "Couldn't get appliance IP address.")
		output_xml_for_error(error_msg)
		sys.exit(2)
	dbg_msg("dpu ip_addrs: %s" % (client_ip_addr_list,) )


	# if not given metdata.xml, then just cleanup
	# if given metadata.xml, then attach disks to the vm
	if options.xml_filename != None and options.xml_filename != "none":
		# get metadata.xml
		f = open(options.xml_filename, "r")
		# This try except is from
		# https://stackoverflow.com/questions/34143685/correct-encoding-for-xml-file-read-from-stdin
		try:
			input = f.buffer
		except AttributeError:
			input = f
		xmlString = input.read()
		##doc = xml.dom.minidom.parseString(xmlString)

		metadata_et = ElementTree.fromstring(xmlString)
		#print "%s : %s : %s\n" % (tree.tag, tree.attrib, tree.text)
		#for child in tree:  
		#	dbg_msg("%s : %s\n" % (child.tag, child.text) )
		#	print "%s : %s" % (child.tag, child.text)

		# get vm info
		vm_uuid = options.uuid
		vm_info, error_msg = ntnx_vm_get_info(vm_uuid)
		if (error_msg != None):
			error_msg = append_error_msg( error_msg, "Couldn't get VM information from Nutanix cluster.")
			output_xml_for_error(error_msg)
			sys.exit(2)

		prior_disks = metadata_et.find('disks')
		if prior_disks == None:
			error_msg = "missing disks in xml file"
			output_xml_for_error(error_msg)
			sys.exit(2)

		diskindexmap = []
		prior_vm_disks = prep_vm_disks_from_et(prior_disks, diskindexmap)

		# if no disks, then don't clone from volume group and attaching with iscsi 
		if (len(prior_disks) > 0) and (len(prior_vm_disks) > 0) and (options.volumegroup != "none"):

			# get vg info
			vg_uuid = options.volumegroup
			(vg_info, error_msg) = ntnx_get_volume_group(vg_uuid)
			if (error_msg != None):
				dbg_msg("Could not get volume group: %s" % vg_uuid)
				output_xml_for_error(error_msg)
				sys.exit(2)

			# attach cdrom
			#cdrom_disk_info = prep_cdrom_disk_from_prior_vm_disks(prior_vm_disks)
			#if cdrom_disk_info:
			#	result, error_msg = ntnx_vm_update_disk_info(vm_uuid, cdrom_disk_info)
			#	if error_msg != None:
			#		dbg_msg("Could not update vm disk info")

			# prep to attach disks
			vm_disks, error_msg =  prep_vm_disks_for_attach(prior_vm_disks, diskindexmap, vg_info)
			if (error_msg != None):
				dbg_msg("Could not prep vm disks for attach to vm %s" % error_msg)
				error_msg = append_error_msg( error_msg, "Couldn't prepare disk information for attaching to VM.")
				output_xml_for_error(error_msg)
				sys.exit(2)

			# attach clone of vg disks to vm
			result, error_msg = ntnx_vm_attach_disks(vm_uuid, vm_disks)
			if (error_msg != None):
				dbg_msg("Could not attach disks to vm")
				error_msg = append_error_msg( error_msg, "Couldn't attach disks to VM.")
				output_xml_for_error(error_msg)
				sys.exit(2)

			# update boot field.  this can only be done after adding the disks to the vm.
			boot_et = metadata_et.find('boot')
			if boot_et != None:
				bootvals = prep_boot_info_from_et(boot_et)
				result, error_msg = ntnx_vm_update_boot(vm_uuid, bootvals)
				if error_msg != None:
					dbg_msg("Could not update vm boot info")
					# TODO: set warning or info msg

	if options.volumegroup != "none":
		# cleanup iscsi
		cleanup_iscsi( target_ip_addr, vm_uuid )
		remove_teardown_entry_iscsi( target_ip_addr, vm_uuid )

		# cleanup vg
		cleanup_vg( client_ip_addr_list, vm_uuid )
		if vg_uuid != None:
			remove_teardown_entry_vg_uuid( vg_uuid )

	# remove vm_restoring entry to info file
	remove_teardown_entry_vm_restoring( vm_uuid )

	# build xml result
	root_aprotect = Element('aprotect')
	child_aprotect = SubElement(root_aprotect, 'postrestore_result')
	child_aprotect_error_msg = SubElement(child_aprotect, 'status')
	child_aprotect_error_msg.text = 'success'

	# dump xml
	indent(root_aprotect)
	ElementTree.dump(root_aprotect)

	dbg_msg("Leaving do_postrestore")

def do_cli_usage(isShowFullUsage=0):
	cmd = os.path.basename( sys.argv[0] )
	print "Usage: %s --server server --cli command [key=value ...]" % cmd
	print "       %s --server server --cli help" % cmd
	if not isShowFullUsage:
		exit(1)
	print ""
	print "where"
	print "  server     Nutanix cluster IP address"
	print ""
	print "Commands:"
	print ""
	print "  ss.list, snapshot.list, snap.list         List snapshots"
	print "  ss.remove, snapshot.remove, snap.remove   Remove snapshots"
	print "  vg.list, volumegroup.list                 List volume groups"
	print "  vm.list                                   List virtual machines"
	print "  help                                      Display this help"
	print ""
	print "Options:"
	print ""
	print "  match=string                 match any ame or uuid containing string "
	print "  name=string                  match name of object"
	print "  vm_name=string               match vm_name"
	print "  vm_uuid=string               match vm_uuid, VM UUID"
	print "  ss_uuid=string               match ss_uuid, snapshot UUID"
	print "  vg_uuid=string               match vg_uuid, volume group UUID"
	print "  alluni                       match unitrends snapshots"
	print "  allother                     match other than Unitrends snapshots"
	print "  force                        auto answer yes to remove snapshots"
	print ""
	print "Examples:"
	print ""
	print "  # list all VMs"
	print "  aprotect.py --server 192.168.171.240 --cli vm.list"
	print ""
	print "  # list VMs with name or UUID matching substring"
	print "  aprotect.py --server 192.168.171.240 --cli vm.list match=ubuntu"
	print ""
	print "  # list VM with specificied name"
	print "  aprotect.py --server 192.168.171.240 --cli vm.list name=pm-ubuntu"
	print ""
	print "  # list snapshots for specified VM UUID"
	print "  aprotect.py --server 192.168.171.240 --cli ss.list vm_uuid=84991b70-866d-4ab2-aec8-06892e145809"
	print ""
	print "  # list snapshots for specified VM UUID and matching substring"
	print "  aprotect.py --server 192.168.171.240 --cli ss.list vm_uuid=ee76d510-fa15-4957-89fc-d1bcc717506d match=171.252"
	print ""
	print "  # list snapshots with name or uuid matching substring"
	print "  aprotect.py --server 192.168.171.240 --cli ss.list match=171.252"
	exit(1)

def do_cli_vm_list(index):
	vm_uuid = None
	match = None
	match_name = None
	match_vm_uuid = None

	index += 1
	for index in range(index, len(sys.argv)):
		arg = sys.argv[index]
		if '=' in arg:
			parts = arg.split('=')
			if (parts > 1):
				key = parts[0]
				value = parts[1]
				if key in ['name', 'vm_name']:
					match_name = value
				elif key in ['vm_uuid']:
					match_vm_uuid = value
				elif key in ['match']:
					match = value			
				else:
					error_msg = "Invalid key: %s" % key
					print error_msg
					do_cli_usage()
		else:
			print "Invalid option: %s" % arg
			do_cli_usage()

	(result, error_msg) = ntnx_vm_get(match_vm_uuid)
	if error_msg:
		dbg_msg(error_msg)
		print "Failed list VMs"
		return

	if match_vm_uuid:
		number_of_vms = 1
		entities = [ result ]
	else:
		number_of_vms = result['metadata']['count']
		entities = result['entities']

	print "%-40s %-36s" % ('vm_name', 'vm_uuid')
	for entity in entities:
		vm_name = entity['name']
		vm_uuid = entity['uuid']
		if match and not ((match in vm_name) or (match in vm_uuid)):
			continue
		if match_name and (match_name != vm_name):
			continue
		if match_vm_uuid and (match_vm_uuid != vm_uuid):
			continue
		print "%-40s %-36s" % (vm_name, vm_uuid)

	return

def do_cli_ss_list(index):
	vm_uuid = None
	match = None
	match_name = None
	match_ss_uuid = None
	match_alluni = None
	match_allother = None
	index += 1
	for index in range(index, len(sys.argv)):
		arg = sys.argv[index]
		if '=' in arg:
			parts = arg.split('=')
			if (parts > 1):
				key = parts[0]
				value = parts[1]
				if key in ['name', 'ss_name']:
					match_name = value
				elif key in ['ss_uuid']:
					match_ss_uuid = value
				elif key in ['vm_uuid']:
					vm_uuid = value
				elif key in ['match']:
					match = value			
				else:
					error_msg = "Invalid key: %s" % key
					print error_msg
					do_cli_usage()
		elif arg in ['alluni']:
			match_alluni = True
		elif arg in ['allother']:
			match_allother = True
		else:
			print "Invalid option: %s" % arg
			do_cli_usage()

	(ss_list_result, error_msg) = ntnx_ss_list(vm_uuid)
	if error_msg:
		dbg_msg(error_msg)
		print "Failed list snapshots"
		return

	total_matches = ss_list_result['metadata']['total_matches']
	dbg_msg("found %d snapshots for vm_uuid %s" % (total_matches, vm_uuid))

	entities = ss_list_result['entities']
	print "%-60s %-36s %-9s %s" % ('ss_name', 'ss_uuid', 'state', 'snapshot_type')
	for entity in entities:
		kind = entity['metadata']['kind']
		ss_uuid = entity['metadata']['uuid']
		name = entity['status']['name']
		state = entity['status']['state']
		snapshot_type = entity['status']['snapshot_type']
		if match_alluni and not name.startswith('ub:') and not match_allother:
			continue
		if match_allother and name.startswith('ub:') and not match_alluni:
			continue
		if match and not ((match in name) or (match in ss_uuid)):
			continue
		if match_name and (match_name != name):
			continue
		if match_ss_uuid and (match_ss_uuid != ss_uuid):
			continue
		print "%-60s %-36s %-9s %s" % (name, ss_uuid, state, snapshot_type)

	return

def do_cli_ss_remove(index):
	vm_uuid = None
	match_ss_uuid = None
	match_name = None
	match = None
	match_alluni = None
	match_allother = None
	arg_force = False
	index += 1
	for index in range(index, len(sys.argv)):
		arg = sys.argv[index]
		if '=' in arg:
			parts = arg.split('=')
			if (parts > 1):
				key = parts[0]
				value = parts[1]
				#if key in 'uuid':
				if key in ['vm_uuid']:
					vm_uuid = value
				elif key in ['ss_uuid']:
					match_ss_uuid = value
				if key in ['name', 'ss_name']:
					match_name = value
				elif key in ['match']:
					match = value			
				else:
					error_msg = "Invalid key: %s" % key
					print error_msg
					do_cli_usage()
		elif arg in ['alluni']:
			match_alluni = True
		elif arg in ['allother']:
			match_allother = True
		elif arg in ['force']:
			arg_force = True
		else:
			print "Invalid option: %s" % arg
			do_cli_usage()

	if not match_ss_uuid and not match_name and not match and not match_alluni and not match_allother:
		print 'No search criteria specified.'
		do_cli_usage()

	(ss_list_result, error_msg) = ntnx_ss_list(vm_uuid)
	if error_msg:
		dbg_msg(error_msg)
		print "Failed list snapshots"
		return

	total_matches = ss_list_result['metadata']['total_matches']
	dbg_msg("found %d snapshots for vm_uuid %s" % (total_matches, vm_uuid))

	entities = ss_list_result['entities']
	print "%-60s %-36s %-9s" % ('ss_name', 'ss_uuid', 'state')
	for entity in entities:
		kind = entity['metadata']['kind']
		ss_uuid = entity['metadata']['uuid']
		name = entity['status']['name']
		state = entity['status']['state']
		snapshot_type = entity['status']['snapshot_type']

		if match_alluni and not name.startswith('ub:') and not match_allother:
			continue
		if match_allother and name.startswith('ub:') and not match_alluni:
			continue
		if match and not ((match in name) or (match in ss_uuid)):
			continue
		if match_ss_uuid and (match_ss_uuid != ss_uuid):
			continue
		if match_name and (match_name != name):
			continue

		print "%-60s %-36s %-9s" % (name, ss_uuid, state)

		try:
			if arg_force:
				answer = 'yes'
			else:
				answer = raw_input( "Remove snapshot name='%s'? " % (name) )
				while answer and answer in ['y', 'ye', 'Y', 'YE', 'YES']:
					answer = raw_input('Type "yes" to remove snapshot: ')
			if answer != None and answer == "yes":
				print "Removing snapshot %s" % (name)
				(result, error_msg) = cleanup_snapshot_by_ss_uuid(ss_uuid)
				if (error_msg != None):
					print("Failed to remove snapshot: " + error_msg)
					dbg_msg(error_msg)
		except KeyboardInterrupt:
			print "\nInterrupted"
			break
		except EOFError:
			print
			break
	
	return

def do_cli_vg_list(index):
	vm_uuid = None
	match = None
	match_vg_uuid = None
	match_vg_name = None
	index += 1
	for index in range(index, len(sys.argv)):
		arg = sys.argv[index]
		if '=' in arg:
			parts = arg.split('=')
			if (parts > 1):
				key = parts[0]
				value = parts[1]
				if key == 'vg_uuid':
					match_vg_uuid = value
				elif key == 'match':
					match = value			
				else:
					error_msg = "Invalid key: %s" % key
					print error_msg
					do_cli_usage()
		else:
			print "Invalid option: %s" % arg
			do_cli_usage()

	(result, error_msg) = ntnx_get_volume_groups()
	if error_msg:
		dbg_msg(error_msg)
		print "Failed list volume groups"
		return

	count = result['metadata']['count']
	dbg_msg("found %d volumegroups" % count)

	print "%-40s %-36s %s" % ('vg_name', 'vg_uuid', 'vg_desc')
	entities = result['entities']
	for index in range(0,count):
		entity = entities[index]

		vg_name = entity['name']
		vg_uuid = entity['uuid']
		vg_desc = entity.get('description', "")

		if match and not ((match in vg_name) or (match in vg_uuid)):
			continue
		if match_vg_uuid and (match_vg_uuid != vg_uuid):
			continue
		if match_vg_name and (match_vg_name != vg_name):
			continue

		print "%-40s %-36s %s" % (vg_name, vg_uuid, vg_desc)

def do_cli():
	#print "This is the name of the script: ", sys.argv[0]
	#print "Number of arguments: ", len(sys.argv)
	#print "The arguments are: " , str(sys.argv)

	# don't retry Server5xx error in CLI
	global gServer5xxRetryCount
	gServer5xxRetryCount = 0

	err = 0

	if options.server == None:
		print "Server not specified."
		do_cli_usage()

	if options.username == None or options.password == None:
		# lookup credentials from database
		result, error_msg = get_server_credentials(options.server)
		if (error_msg != None):
			dbg_msg(error_msg)
			print "Server credentials not found."
			err = 1
		else:
			username, password, domain = result
			if domain != None and len(domain) > 0:
				options.username = username + '@' + domain
			else:
				options.username = username
			options.password = password

	if err:
		return

	# locate --cli option
	for index in range(1, len(sys.argv)):
		arg = sys.argv[index]
		if arg == '--cli':
			index += 1
			break
	#for index in range(index, len(sys.argv)):
	if index >= len(sys.argv):
		do_cli_usage()
	arg = sys.argv[index]
	if arg in ['vm.list']:
		do_cli_vm_list(index)
	elif arg in ['ss.list', 'snap.list', 'snapshot.list']:
		do_cli_ss_list(index)
	elif arg in ['ss.remove', 'snap.remove', 'snapshot.remove']:
		do_cli_ss_remove(index)
	elif arg in ['vg.list', 'volumegroup.list']:
		do_cli_vg_list(index)
	elif arg in ['help']:
		do_cli_usage(1)
	else:
		error_msg = "Error: Invalid cli command: %s" % arg
		print error_msg
		do_cli_usage()
			
	return

def validate_args(options):

	error_msg = None

	if (options.cleanupidle == None) and (options.cli == None):
		if ( (options.server == None) or
			(options.username == None) or
			(options.password == None) ):
			error_msg = "Server, Username and Password are mandatory."
			output_xml_for_error(error_msg)
			sys.exit(2)

	# Backup & Restore mode checks will go here
	if ( (options.bpcmd == None) and
		(options.disklist == None) and
		(options.backup == None) and
		(options.postbackup == None) and
		(options.restore == None) and
		(options.postrestore == None) and
		(options.cleanupidle == None) and
		(options.cli == None) and
		(options.prmtest == None)):
		error_msg = "Invalid mode for operation passed."
		output_xml_for_error(error_msg) 
		sys.exit(2)

	if (options.restore):
		if (options.newvmname == None):
			error_msg = "Restore mode requires newvmname option."
		elif (options.xml_filename == None):
			error_msg = "Restore mode requires xml_filename option."
		elif (options.storagecontainer == None):
			error_msg = "Restore mode requires storagecontainer option."
		if error_msg:
			output_xml_for_error(error_msg) 
			sys.exit(2)

	if (options.postrestore):
		if (options.xml_filename == None):
			error_msg = "PostRestore mode requires xml_filename option."
		elif (options.uuid == None):
			error_msg = "PostRestore mode requires uuid option."
		elif (options.volumegroup == None):
			error_msg = "PostRestore mode requires volumegroup option."
		if error_msg:
			output_xml_for_error(error_msg) 
			sys.exit(2)

	# validate quiesce setting
	if  options.quiesce != None and not options.quiesce.upper() in ['YES', 'NO']:
		error_msg = "Invalid value specified for quiesce option"
		output_xml_for_error(error_msg) 
		sys.exit(2)

	# validate metadata setting
	if  options.metadata != None and not options.metadata.upper() in ['YES', 'NO']:
		error_msg = "Invalid value specified for metadata option"
		output_xml_for_error(error_msg) 
		sys.exit(2)

	if options.ppid == None:
		options.ppid = str( os.getppid() ) ##"0"

def prep_ini_values():
	global gServer5xxRetryCount
	global gTaskWaitTime
	global gTaskRetyCount
	global gIscsiCmdRetryCount
	global gTraceLevel

	gServer5xxRetryCount = ini_get_int("aprotect", "Server5xxRetryCount", 5)
	gTaskWaitTime = ini_get_int("aprotect", "TaskWaitTime", 60)
	gTaskRetryCount = ini_get_int("aprotect", "TaskRetryCount", 5)
	gIscsiCmdRetryCount = ini_get_int("aprotect", "IscsiCmdRetryCount", 5)
	gTraceLevel = ini_get_int("debugging", "aprotect.py", 1)

def main():
	global gMsgLock
	global gVerbose
	gMsgLock = Lock()

	if options.cli and not (('-v' in sys.argv) or ('--verbose' in sys.argv)):
		gVerbose = 0

	dbg_msg("Running aprotect's PY interface..")

	validate_args(options)
	prep_ini_values()

	dbg_msg("IP Address : %s" % options.server)
	dbg_msg("Username   : %s" % options.username)
	# TODO: remove this devtest workaround
	# if options.username != None and '\\' in options.username:
	#	fields = options.username.split('\\')
	#	options.username = fields[1] + '@' + fields[0]
	# TODO : Remove this later
	##dbg_msg("Password   : %s" % options.password)
	dbg_msg("Password   : %s" % "*")
	dbg_msg("bpcmd is   : %s" % options.bpcmd)
	dbg_msg("Disklist   : %s" % options.disklist)
	dbg_msg("Backup     : %s" % options.backup)
	dbg_msg("UUID       : %s" % options.uuid)
	dbg_msg("Postbackup : %s" % options.postbackup)
	dbg_msg("Snapshot   : %s" % options.snapshot)
	dbg_msg("volumegroup: %s" % options.volumegroup)
	dbg_msg("hardwarehash: %s" % options.hardwarehash)
	dbg_msg("metadata   : %s" % options.metadata)

	# UNIBP-19306: allow insecure cert if CentOS7 
	if hasattr(ssl, "_https_verify_certificates"):
		ssl._https_verify_certificates(False)

	# Put all of the main execution code in a "try..except".
	# Helpful if any unhandled errors occur.
	try:
		if  (options.bpcmd != None):
			dbg_msg("Executing bpcmd for %s" % options.bpcmd)
			error_msg = get_xml_for_bpcmd()
			if (error_msg != None):
				output_xml_for_error(error_msg)
				sys.exit(2)
			else:
				dbg_msg("Successfully retrieved information for %s" % options.bpcmd)
		elif (options.disklist == 1):
			dbg_msg("Executing disk listing command for VM %s" % options.uuid)
			error_msg = get_disk_info_for_vm()
			if (error_msg != None):
				output_xml_for_error(error_msg)
				sys.exit(2)
			else:
				dbg_msg("Successfully retrieved information for disk listing for VM %s" \
					% options.uuid)
		elif (options.backup):
			dbg_msg("Executing backup for VM %s" % options.uuid)
			error_msg = do_backup()
			if (error_msg != None):
				output_xml_for_error(error_msg)
				sys.exit(2)
			else:
				dbg_msg("Successfully prepared for backup of VM %s" \
					% options.uuid)

		elif (options.postbackup):
			dbg_msg("Executing postbackup for VM %s" % options.uuid)
			signal.signal(signal.SIGTERM, sig_handler)
			error_msg = do_postbackup()
			if (error_msg != None):
				output_xml_for_error(error_msg)
				sys.exit(2)
			else:
				dbg_msg("Successfully did postbackup of VM %s" \
					% options.uuid)

		elif (options.restore):
			dbg_msg("Executing restore")
			error_msg = do_restore()
			if (error_msg != None):
				output_xml_for_error(error_msg)
				sys.exit(2)
			else:
				dbg_msg("Successfully did restore ")

		elif (options.postrestore):
			dbg_msg("Executing postrestore")
			signal.signal(signal.SIGTERM, sig_handler)
			error_msg = do_postrestore()
			if (error_msg != None):
				output_xml_for_error(error_msg)
				sys.exit(2)
			else:
				dbg_msg("Successfully did postrestore ")

		elif (options.cleanupidle):
			process_idle_info_files() 

		elif (options.cli):
			do_cli()

		elif (options.prmtest):
			doprmtest()
		else:
			dbg_msg("Unsupported command")
			error_msg = "Unsupported command"
			output_xml_for_error(error_msg)
			sys.exit(2)

	except Exception as e:
		ex_string = traceback.format_exc()
            	dbg_msg("Error:\n%s" % ex_string)
		if isinstance(e, MemoryError):
			error_msg = "Internal error: Out of memory"
		elif isinstance(e, OSError):
			error_msg = "Internal error: %s" % (e)
		else:
			error_msg = "Internal error"
		dbg_msg(error_msg)
		output_xml_for_error(error_msg)
		sys.exit(2)

if __name__ == "__main__":
	main()

